'''
This is a series of utility functions for TransitFit, including input validation
'''
import pandas as pd
import random
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

import numpy as np
from .lightcurve import LightCurve
from collections.abc import Iterable
import os

def validate_variable_key(key):
    '''
    Checks that a key is valid for use with PriorInfo, and corrects when it
    is obvious what is meant. Raises KeyError if unable to correct.

    '''
    if key.lower() in ['p','period']:
        return 'P'
    if key.lower() in ['rp','r_p','radius','planet_radius', 'planet radius']:
        return 'rp'
    if key.lower() in ['a', 'semimajor axis', 'semimajor_axis']:
        return 'a'
    if key.lower() in ['inc','inclination']:
        return 'inc'
    if key.lower() in ['t0', 't_0']:
        return 't0'
    if key.lower() in ['ecc','e','eccentricity']:
        return 'ecc'
    if key.lower() in ['w', 'periastron', 'longitude_of_periastron', 'longitude of periastron']:
        return 'w'
    if key.lower() in ['q0']:
        return 'q0'
    if key.lower() in ['q1']:
        return 'q1'
    if key.lower() in ['q2']:
        return 'q2'
    if key.lower() in ['q3']:
        return 'q3'

    raise KeyError('Unable to recognise variable name {}'.format(key))


def calculate_logg(host_mass, host_radius):
    '''
    Calculates log10(g) for a host in units usable by TransitFit.
    g is in cm s^-2

    Parameters
    ----------
    host_mass : tuple
        The host mass in solar masses and the uncertainty
    host_radius : tuple
        The host radius in solar radii and the uncertainty
    '''
    m_sun = 1.989e30
    r_sun = 6.957e8
    G = 6.674e-11

    m = host_mass[0] * m_sun
    err_m = host_mass[1] * m_sun
    r = host_radius[0] * r_sun
    err_r = host_radius[1] * r_sun

    g = m * G * r ** -2 * 100 # factor 100 converts to cm s^-2
    err_g = G/(r ** 2) * np.sqrt(err_m ** 2 + (4 * m**2 * err_r ** 2)/r**2) * 100

    logg = np.log10(g)
    err_logg = err_g / (g * np.log(10))

    return logg, err_logg


def get_normalised_weights(results):
    '''
    Obtains normalised weights for the Dynesty results

    Parameters
    ----------
    results : dynesty.results.Results
        The Dynesty results object

    Returns
    -------
    weights : np.array, shape (n_iterations,)
        The normalised weights for each sample set
    '''
    return np.exp(results.logwt - results.logwt.max())/np.sum(np.exp(results.logwt - results.logwt.max()))


def get_covariance_matrix(results):
    '''
    Gets a covariance matrix from Dynesty results

    Parameters
    ----------
    results : dynesty.results.Results
        The Dynesty results object, but must also have weights as an entry

    Returns
    -------
    cov : np.array, shape (ndims, ndims)
        The covariance matrix for the results object.
    '''
    # Calculate a covariance matrix using numpy
    cov = np.cov(results.samples, rowvar=False, aweights=results.weights)

    return cov


def weighted_avg_and_std(values, weights, axis=-1, single_val=False):
    '''
    Calculates the weighted average and error on some data.

    axis defaults to the last one (-1)
    '''
    if not isinstance(values, Iterable):
        values = [values]
    if not isinstance(weights, Iterable):
        weights = [weights]

    values = np.array(values)
    weights = np.array(weights)

    if not isinstance(values[0], Iterable):
        single_val = True

    if single_val:
        # Flatten the values and weights
        try:
            flat_vals = np.array([i for epoch_vals in values for i in epoch_vals])
            flat_weights = np.array([i for epoch_vals in weights for i in epoch_vals])
        except:
            flat_vals = values
            flat_weights = weights
        average = np.average(flat_vals, weights=1/(flat_weights**2))
        uncertainty = 1 / np.sqrt(np.sum(1/(flat_weights**2)))

        return average, uncertainty

    # Make blank arrays to loop over the entries in values and weights
    average = []
    uncertainty = []

    for i in range(len(values)):
        average.append(np.average(np.array(values[i]), weights=1/(np.array(weights[i])**2)))
        uncertainty.append(1 / np.sqrt(np.sum(1/(np.array(weights[i])**2))))

    return np.array(average), np.array(uncertainty)


def AU_to_host_radii(a, R, a_err=0, R_err=0, calc_err=False):
    '''
    Converts a number in AU to a value in host radii when given the host
    radius R in Solar radii. Inverse of host_radii_to_AU.
    '''
    AU = 1.495978707e11
    R_sun = 6.957e8

    if not calc_err:
        return (a * AU) / (R * R_sun)

    err = np.sqrt( ((R_err * R_sun) * (a*AU)/((R*R_sun)**2)) ** 2 + ((a_err * AU)/(R*R_sun))**2  )

    return (a * AU) / (R * R_sun), err/(R * R_sun)


def host_radii_to_AU(a, R, a_err=0, R_err=0, calc_err=False):
    '''
    converts a separation in host radii into a separation in AU when given
    the host radius R in Solar radii. Inverse of AU_to_host_radii.
    '''
    AU = 1.495978707e11
    R_sun = 6.957e8

    if not calc_err:
        return (a * R * R_sun)/AU

    err = np.sqrt ((R * R_sun * a_err )**2 + (a * R_err * R_sun)**2)

    return (a * R * R_sun)/AU, err/AU


def estimate_t14(Rp, Rs, a, P):
    '''
    Estimates t14 in minutes, if P is in days
    '''
    AU = 1.495978707e11
    R_sun = 6.957e8
    R_jup = 71492000

    return (Rp * R_jup + Rs * R_sun)/(np.pi * a * AU) * P * 24 * 60

def split_lightcurve_file(path, t0, P,t14=20, cutoff=0.25, window=5,
                          new_base_fname='split_curve'):
    '''
    Split a light curve file into multiple single epoch files

    Splits a multi-epoch lightcurve data file into multiple single-epoch files
    and saves these. This is useful for dealing with multi-epoch observations
    which contain TTVs, or have long term periodic trends, since single-epoch
    observation trends can be approximated with polynomial fits. New files
    are created of the form new_base_name_X

    Parameters
    ----------
    path : str
        The path to the light curve data to be split
    t0 : float
        The centre of a transit
    P : float
        The estimated period of the planet in days
    t14 : float, optional
        The approximate transit duration in minutes. Default is 20
    cutoff : float, optional
        If there are no data within t14 * cutoff of t0, a period will be
        discarded. Default is 0.25
    window : float, optional
        Data outside of the range [t0 ± (0.5 * t14) * window] will be
        discarded. Default is 5.
    new_base_fname : str, optional
        The base name for the new files, which will have numbers appended
        depending on the epoch. This can be used to specify a relative path
        for saving. Default is `'split_curve'`.
    return_names : bool


    Returns
    -------
    paths : array_like, shape (n_curves,)
        The paths to each of the new data files
    '''
    from .io import read_data_file

    # Load the full data file as a LightCurve
    full_lightcurve = LightCurve(*read_data_file(path))

    # Split the full curve into individual epochs
    single_epoch_curves = full_lightcurve.split(t0, P, t14, cutoff, window)
    dirname = os.path.dirname(path)

    paths = []

    # Now save all the new curves
    for i, curve in enumerate(single_epoch_curves):
        fname = new_base_fname + '_{}'.format(i)
        curve.save(os.path.join(dirname, fname))
        paths.append(os.path.join(dirname, fname))

    return paths


def check_batches(allow_ttv, data_files):
    """Taken from firefly/_utils.py, with minor modifications.
    This was needed to prevent uneven batchsizes.
    Now it is not required in most of the cases, but has been kept as a failsafe.

    Args:
        allow_ttv (bool): To allow for TTV or not.
        data_files (str): The path to the data input .csv file, which contains the paths to the
        light curve data.

    Returns:
        str:  The path to modified data input .csv file, which contains the paths to the
        light curve data. It removed some lightcurves randomly to take care of batching issue in uneven batches.
    """

    print('Checking if batches can be made of equal size.')
    from natsort import natsorted
    df = pd.read_csv(data_files)
    light_curve_paths = df['Path'].to_list()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Enforce batches be the same size
    curves = len(light_curve_paths)
    if int(curves) > 5:
        print('Still checking...')
        if allow_ttv == True:
            equal_batches = [5 + 3 * n for n in range(0, 500)]
        else:
            equal_batches = [6 + 4 * n for n in range(0, 500)]

        if int(curves) not in equal_batches:
            import bisect

            index = bisect.bisect(equal_batches, int(curves))
            new_curves = equal_batches[index - 1]
            print(
                f"\nEnforcing lightcurves to be in equal batch sizes. Discarded {curves-new_curves}."
            )
            curves = new_curves
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
    # Sort the files into ascending order and take random sample
    if len(light_curve_paths) == int(curves):
        return data_files

    light_curve_paths = random.sample(light_curve_paths, k=int(curves))
    light_curve_paths = natsorted(light_curve_paths)
    print(
        f"\nA total of {len(light_curve_paths)}",
        "lightcurves" if len(light_curve_paths) > 1 else "lightcurve",
        "will be used.",
    )

    path_temp = os.path.dirname(os.path.abspath(data_files))
    data_path = f"{path_temp}/data_paths_modified.csv"
    df = df.loc[df['Path'].isin(light_curve_paths)]
    df.to_csv(data_path, index=False, header=True)
    print(f"New data path is {data_path}")

    return data_path



'''
Module to deal with all the reading in of inputs, as well as printing final results to terminal
'''

import numpy as np
import pandas as pd
from .priorinfo import PriorInfo, _prior_info_defaults, setup_priors
import dynesty
import csv
import os
from ._utils import validate_variable_key, AU_to_host_radii
from .lightcurve import LightCurve
import batman
import os

#############################################################
#                         PRIORS                            #
#############################################################
def read_priors_file(path, n_telescopes, n_filters, n_epochs,
                     limb_dark='quadratic', filter_indices=None, folded=False,
                     folded_P=None, folded_t0=None, host_radius=None,
                     allow_ttv=None, lightcurves=None, suppress_warnings=False, 
                     error_scaling=False):
    '''
    If given a csv file containing priors, will produce a PriorInfo object
    based off the given values

    Parameters
    ----------
    path : str
        Path to .csv file containing the priors.

        Columns should me in the order
        --------------------------------------------------------
        |  key  |   mode  |  input_A  |   input_B   |  filter  |
        --------------------------------------------------------

        The available modes and the expected values of inputs A and B are:

        - 'uniform': input A should be lower limit, input B should be upper
                     limit
        - 'gaussian': input_A should be mean, input_B should be standard
                      deviation.
        - 'fixed': input_A should be the fixed value. input_B is not used and
                   should be left blank.

    n_telescopes : int
        The number of different telescopes being used. Required so that
        simultaneous observations from different observatories can be used by
        TransitFit
    n_filters : int
        The number of different filters being used
    n_epochs : int
        The number of different epochs being used
    limb_dark : str, optional
        The model of limb darkening you want to use. Accepted are
            - linear
            - quadratic
            - squareroot
            - power2
            - nonlinear
        Default is quadratic
    filter_indices : array_like, optional
        If provided, will only initialise fitting for parameters which are
        relevant to the filter indices given. Note that this will result in
        a difference between the filter indices used at the top, user level and
        those used within this PriorInfo
    folded : bool, optional
        If True, will not initialise P or t0 from the priors list. Instead will
        use folded_P and folded_t0 to set fixed values. Default is False
    folded_P : float, optional
        Required if folded is True. This is the period that the light curves
        are folded to
    folded_t0 : float, optional
        Required if folded is True. This is the t0 that the light curves are
        folded to
    host_radius : float, optional
        The host radius in Solar radii.
        If this is provided, then will assume that the orbital separation is
        given in AU rather than host radii and will convert the values
        accordingly.

    Notes
    -----
    Detrending currently cannot be initialised in the prior file. It will be
    available as a kwarg in the pipeline function
    '''
    priors_list = pd.read_csv(path).values

    return parse_priors_list(priors_list, n_telescopes, n_filters, n_epochs, limb_dark, filter_indices, folded, folded_P, folded_t0, host_radius, allow_ttv, lightcurves, suppress_warnings, error_scaling)

def parse_priors_list(priors_list, n_telescopes, n_filters,
                      n_epochs, ld_model, filter_indices=None, folded=False,
                      folded_P=None, folded_t0=None, host_radius=None,
                      allow_ttv=False, lightcurves=None, suppress_warnings=False,
                      error_scaling=False):
    '''
    Parses a list of priors to produce a PriorInfo with all fitting parameters
    initialised.

    Parameters
    ----------
    priors_list : array_like, shape(X, 5)
        A list of prior information for each variable to be fitted. Can also
        set fixed values by setting the low and high values to `None` or
        `np.nan`. Each row should be of the form
        [key, mode, input A, input B, filter index]
    n_telescopes : int
        The number of different telescopes being used. Required so that
        simultaneous observations from different observatories can be used by
        TransitFit
    n_filters : int
        The number of different filters being used
    n_epochs : int
        The number of different epochs being used
    ld_model : str
        The limb darkening model to use
    filter_indices : array_like, optional
        If provided, will only initialise fitting for parameters which are
        relevant to the filter indices given. Note that this will result in
        a difference between the filter indices used at the top, user level and
        those used within this PriorInfo
    folded : bool, optional
        If True, will not initialise P or t0 from the priors list. Instead will
        use folded_P and folded_t0 to set fixed values. Default is False
    folded_P : float, optional
        Required if folded is True. This is the period that the light curves
        are folded to
    folded_t0 : float, optional
        Required if folded is True. This is the t0 that the light curves are
        folded to
    host_radius : float, optional
        The host radius in Solar radii. If this is provided, then will assume
        that the orbital separation is given in AU rather than host radii and
        will convert the values accordingly

    Returns
    -------
    priors : PriorInfo
        The fully initialised PriorInfo which can then be used in fitting.
    '''
    if folded:
        if folded_P is None:
            raise ValueError('folded_P must be provided for folded prior mode')
        if folded_t0 is None:
            raise ValueError('folded_t0 must be provided for folded prior mode')

    if allow_ttv and lightcurves is None:
        raise ValueError('lightcurves must be provided if allow_ttv is True')

    # Make a dictionary of the defaults
    priors_dict = {}

    if filter_indices is None:
        # We will use all filters
        filter_indices = np.arange(n_filters)

    rp_count = 0
    used_filters = []
    for row in priors_list:
        # First check the key and correct if possible
        row[0] = validate_variable_key(row[0])

        # Now add to the priors_dict
        if row[0] in ['rp']:
            # We have to deal with extracting particular filters

            if row[-1] in filter_indices:
                priors_dict[row[0]] = np.append(row[2:-1], rp_count)
                rp_count += 1
                used_filters.append(row[-1])

        else:
            if row[0] == 'a' and host_radius is not None:
                # Convert the inputs to host radius units
                row[2] = AU_to_host_radii(row[2], host_radius)
                row[3] = AU_to_host_radii(row[3], host_radius)

            priors_dict[row[0]] = row[2:]

    # We  have to convert between the global filter indexing and an
    # internal filter indexing here.
    filter_conversion = {ai : i for i, ai in enumerate(used_filters)}
    # Each key is the global value, and converts to the internal value.

    ##############################
    # Make the default PriorInfo #
    ##############################

    # Need to check if any variables are missing from the default prior
    for key in _prior_info_defaults:
        if key not in priors_dict:
            # Has not been specified in the priors list
            priors_dict[key] = [_prior_info_defaults[key]]

    # Update priors_dict values if using folded curves
    if folded:
        #
        priors_dict['P'] = [folded_P]
        priors_dict['t0'] = [folded_t0]

    # Now make the basic PriorInfo
    priors = setup_priors(priors_dict['P'][0],
                          priors_dict['t0'][0],
                          priors_dict['a'][0],
                          priors_dict['rp'][0],
                          priors_dict['inc'][0],
                          priors_dict['ecc'][0],
                          priors_dict['w'][0],
                          ld_model, n_telescopes, n_filters, n_epochs,
                          priors_dict['q0'][0],
                          priors_dict['q1'][0],
                          priors_dict['q2'][0],
                          priors_dict['q3'][0],
                          allow_ttv, lightcurves, error_scaling)

    ##########################
    # Initialise the fitting #
    ##########################
    for ri, row in enumerate(priors_list):
        key, mode, inputA, inputB, filt = row
        mode = mode.strip()

        if pd.notna(filt) and filt not in filter_indices:
            # Skip this parameter since it's not in the filters we are
            # interested in
            pass

        else:
            # Convert the filter index
            try:
                filt = filter_conversion[filt]
            except KeyError:
                if pd.isna(filt):
                    # This was excepted because filt was not specified
                    filt = None
                else: raise

            if key == 'a' and host_radius is not None:
                # Convert the inputs to host radius units
                row[2] = AU_to_host_radii(row[2], host_radius)
                row[3] = AU_to_host_radii(row[3], host_radius)

            if key in ['P', 't0'] and folded:
                # Skip P and t0 for folded mode - we aren't fitting them
                pass

            elif key in ['P'] and allow_ttv:
                # We can't fit period if we are using ttv mode - skip it
                if not suppress_warnings:
                    print("WARNING: Ignoring P fitting due to ttv mode. It is recommended to specify P as 'Fixed' in the input file, else TransitFit will default to the value given with Input A.")
                pass

            elif mode.lower() in ['fixed', 'f', 'constant', 'c']:
                # Not being fitted. Default value was specified.
                pass

            elif mode.lower() in ['uniform', 'unif', 'u']:
                # Uniform fitting
                if key in ['t0'] and allow_ttv:
                    for li in np.ndindex(lightcurves.shape):
                        # Loop through each lightcurve and initialise t0
                        if lightcurves[li] is not None:
                            priors.add_uniform_fit_param(key, inputA, inputB, li[0], li[1], li[2])
                else:
                    priors.add_uniform_fit_param(key, inputA, inputB, filter_idx=filt)

            elif mode.lower() in ['gaussian', 'gauss', 'normal', 'norm', 'g']:
                # Gaussian fitting
                if key in ['t0'] and allow_ttv:
                    for li in np.ndindex(lightcurves.shape):
                        # Loop through each lightcurve and initialise t0
                        if lightcurves[li] is not None:
                            priors.add_gaussian_fit_param(key, inputA, inputB, li[0], li[1], li[2])
                else:
                    priors.add_gaussian_fit_param(key, inputA, inputB, filter_idx=filt)

            else:
                raise ValueError('Unrecognised fiting mode {} in input row {}. Must be any of "uniform", "gaussian", or "fixed"'.format(mode, ri))

    return priors

#############################################################
#                       DATA FILES                          #
#############################################################
def read_input_file(path, skiprows=0):
    '''
    Reads in a file with listed inputs and produces data arrays for use in
    retrieval.

    Parameters
    ----------
    path : str
        The path to the .csv file with the paths to input parameters and
        their telescope, filter, epoch, and detrending indices.

        Columns should be in the order
        ------------------------------------------------------------
        |  path  |  telescope  |  filter  |  epoch  |  detrending  |
        ------------------------------------------------------------

    Returns
    -------
    lightcurves  : np.array of `LightCurve`s, shape (n_telescopes, n_filters, n_epochs)
        The data loaded and stored as an array of LightCurves
    detrending_index_array : array_like, shape (n_telescopes, n_filters, n_epochs)
        The detrending indices for each lightcurve
    '''
    info = pd.read_csv(path).values

    data_path_array, detrending_index_array = parse_data_path_list(info)

    lightcurves = read_data_path_array(data_path_array, skiprows=skiprows)

    return lightcurves, detrending_index_array

def _read_data_csv(path, usecols=None):
    '''
    Given a path to a csv with columns [time, flux, errors], will get
    all the data in a way which can be used by the Retriever

    '''
    # Read in with pandas
    data = pd.read_csv(path, usecols=usecols)

    # Extract the arrays
    times, flux, errors = data.values.T

    non_nan = np.invert(np.any(pd.isna(data.values.T), axis=0))

    return times[non_nan], flux[non_nan], errors[non_nan]

def _read_data_txt(path, skiprows=0, usecols=None, delimiter=' '):
    '''
    Reads a txt data file with columns
    '''
    try:
        data = pd.read_csv(path, usecols=usecols, dtype=float, delimiter=delimiter)
        times, flux, errors = data.values.T
        non_nan = np.invert(np.any(pd.isna(data.values.T), axis=0))
        return times[non_nan], flux[non_nan], errors[non_nan]

    except Exception as e:
         times, flux, errors = np.loadtxt(path, skiprows=skiprows, usecols=usecols).T
         return times, flux, errors

         raise e

def read_data_file(path, skiprows=0, folder=None, usecols=None, delimiter=None):
    '''
    Reads a file in, assuming that it is either a:
        .csv
        .txt
    with columns in the order time, depth, errors. Note that TransitFit assumes
    BJD times.

    Parameters
    ----------
    path : str
        Full path to the file to be loaded
    skiprows : int, optional
        Number of rows to skip in reading txt file (to avoid headers)
    folder : str or None, optional
        If not None, this folder will be prepended to all the paths. Default is
        None.
    usecols : int or sequence, optional
        Which columns to read, with 0 being the first. For example,
        ``usecols = (1,4,5)`` will extract the 2nd, 5th and 6th columns.
        These should be given in the order time, flux, uncertainty
        The default, None, results in all columns being read.

    Returns
    -------
    times : np.array
        The times of the data series
    flux : np.array
        The flux
    error : np.array
        The uncertainty on the flux
    '''
    if folder is None:
        folder = ''

    if path[-4:] == '.csv':
        times, flux, errors = _read_data_csv(os.path.join(folder, path), usecols)
    elif path[-4:] == '.txt':
        times, flux, errors = _read_data_txt(os.path.join(folder, path), skiprows, usecols)
    else:
        raise ValueError(f'Data files must be .csv or .txt, not {path[-4:]}')

    return times, flux, errors

def read_data_path_array(data_path_array, skiprows=0):
    '''
    If passed an array of paths, will read in to produce an array of
    `LightCurve`s

    Parameters
    ----------
    data_path_array : array_like, shape (n_telescopes, n_filters, n_epochs)
        Array with paths to data to load

    Returns
    -------
    lightcurves  : np.array of `LightCurve`s, shape (n_telescopes, n_filters, n_epochs)
        The data loaded and stored as an array of LightCurves
    '''
    data_path_array = np.array(data_path_array)

    n_telescopes = data_path_array.shape[0]
    n_filters = data_path_array.shape[1]
    n_epochs = data_path_array.shape[2]

    lightcurves = np.full((data_path_array.shape), None, object)

    for i in np.ndindex(lightcurves.shape):
        if data_path_array[i] is not None:
            times, flux, errors = read_data_file(data_path_array[i], skiprows)

            lightcurves[i] = LightCurve(times, flux, errors, i[0], i[1], i[2])

    return lightcurves

def parse_data_path_list(data_path_list):
    '''
    Parses a list of paths to data files and places them into an array which
    can be passed to read_data_file_array

    Parameters
    ----------
    data_path_list : array_like, shape (n_light_curves, 5)
        The list of paths. Each row should contain
        [data path, telescope idx, filter idx, epoch idx, detrending idx]

    Returns
    -------
    data_path_array : array_like, shape (num_filters, num_epochs, num_telescopes)
        The paths inserted into an array which can be used by TransitFit
    detrending_index_array : array_like, shape (num_filters, num_epochs, num_telescopes)
        The array of detrending model indices for each light curve
    '''
    data_path_list = np.array(data_path_list, dtype=object)

    if data_path_list.ndim == 1:
        data_path_list = np.array([data_path_list])

    # Work out how many epochs and filters we have
    n_telescopes = data_path_list[:,1].max() + 1
    n_filters = data_path_list[:,2].max() + 1
    n_epochs = data_path_list[:,3].max() + 1

    n_detrending_models = data_path_list[:,4].max() + 1

    # Initialise a blank array, filling it with None.
    data_path_array = np.full((n_telescopes, n_filters, n_epochs), None, object)
    detrending_index_array = np.full((n_telescopes, n_filters, n_epochs), None, object)
    column_array = np.full((n_telescopes, n_filters, n_epochs), None, object)

    # Populate the blank array
    for row in data_path_list:
        p, i, j, k, l = row
        data_path_array[i, j, k] = p
        detrending_index_array[i, j, k] = l

    return data_path_array, detrending_index_array

#############################################################
#                         FILTERS                           #
#############################################################
def parse_filter_list(filter_list, delimiter=None, unit='nanometers'):
    '''
    Parses a list of filter information into a usable form for the `'filters'`
    argument in PriorInfo.fit_limb_darkening.

    Parameters
    ----------
    filter_list : array_like, shape (n_filters, 3)
        The information on the filters. Each row should contain
        [filter_index, low wavelength/file path, high wavelength]
        The filter indices should refer to the indices used in the priors file.
        Wavelengths should be in nm.
    delimiter : str, optional
        The delimiter used in filter profile files. Default is None, which
        automatically detects using csv.Sniffer

    Returns
    -------
    filter_info : np.array, shape (n_filters, 2)
        The filter information pass to PriorInfo.fit_limb_darkening.
    '''
    filter_list = np.array(filter_list, dtype=object)

    # How many filters are there?
    n_filters = int(filter_list[:,0].max() + 1)

    # Make a blank array to populate with the filter limits
    filter_info = np.zeros((n_filters, 2), object)

    # Now populate!
    for i in range(n_filters):
        fidx, input_A, input_B = filter_list[i]

        # Check if path, name, or limits are provided:
        try:
            input_A = float(input_A)
            input_B = float(input_B)
            limits_provided=True
        except:
            limits_provided=False

        if limits_provided:
            # box limits provided
            filter_info[i] = input_A, input_B
        else:
            # Get either the path from input or replace path for provided filters
            path, unit = get_filter_path(filter_list[i, 1], unit)
            # Load in the filter profile
            filter_profile = pd.read_csv(path, sep=delimiter, dtype=float).values.T

            # Since filter wavelengths need to be in nm, we should convert them
            # from angstroms if necessary:
            if unit.lower() in ['nm', 'nanometers']:
                factor = 1
            elif unit.lower() in ['angstroms', 'angstrom', 'a']:
                factor = 0.1
            elif unit.lower() in ['m','meters','metres']:
                factor = 1e-9
            else:
                raise ValueError(f'Unrecognised unit {unit} given for filter profiles. TransitFit can only recognise "m", "nm", or "angstroms".')
            filter_info[i,0] = filter_profile[0] * factor
            filter_info[i,1] = filter_profile[1]

    return filter_info

def read_filter_info(path, delimiter=None, unit='nanometers'):
    '''
    Reads in information on the filters from .csv file and puts them in a
    format which can be passed to the ``filters`` argument in
    PriorInfo.fit_limb_darkening.

    Parameters
    ----------
    path : str
        Path to the .csv file containing the information on the filters. This
        file should have three columns:

        -------------------------------------------------
        |   filter_idx  |   low_wl_or_path  |   high_wl |
        -------------------------------------------------

        Filters can either be specified as uniform between low_wl and high_wl,
        or a full profile can be passed through a file. Filter profile files
        must be two columns, giving wavelength and transmission fraction in
        range [0,1].
        The filter indices should refer to the indices used in the priors file.
        All wavelengths should be in nm.

    delimiter : str, optional
        The delimiter used in filter profile files. Default is None, which
        automatically detects using csv.Sniffer

    Returns
    -------
    filter_info : np.array, shape (n_filters, 2)
        The filter information pass to PriorInfo.fit_limb_darkening.

    '''

    info = pd.read_csv(path).values

    return parse_filter_list(info, delimiter)

def get_filter_path(input_str, unit):
    '''
    Checks if a path or name of a default provided filter is given.

    Parameters
    ----------
    input : str
        The input string from the filter file input

    Returns
    -------
    path : str
        The path to the relevant filter
    '''
    input_str = input_str.strip()

    if os.path.exists(input_str):
        return input_str, unit

    filter_folder = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    # Johnson UBVRI
    if input_str == 'U':
        return os.path.join(filter_folder, 'filters/JohnsonU.csv'), 'angstroms'
    if input_str == 'B':
        return os.path.join(filter_folder, 'filters/JohnsonB.csv'), 'angstroms'
    if input_str == 'V':
        return os.path.join(filter_folder, 'filters/JohnsonV.csv'), 'angstroms'
    if input_str == 'R':
        return os.path.join(filter_folder, 'filters/CousinsR.csv'), 'angstroms'
    if input_str == 'I':
        return os.path.join(filter_folder, 'filters/CousinsI.csv'), 'angstroms'

    # SDSS filters:
    if input_str == 'g':
        return os.path.join(filter_folder, 'filters/SLOAN_g.csv'), 'angstroms'
    if input_str == 'g_prime':
        return os.path.join(filter_folder, 'filters/SLOAN_gprime.csv'), 'angstroms'
    if input_str == 'i':
        return os.path.join(filter_folder, 'filters/SLOAN_i.csv'), 'angstroms'
    if input_str == 'i_prime':
        return os.path.join(filter_folder, 'filters/SLOAN_iprime.csv'), 'angstroms'
    if input_str == 'r':
        return os.path.join(filter_folder, 'filters/SLOAN_r.csv'), 'angstroms'
    if input_str == 'r_prime':
        return os.path.join(filter_folder, 'filters/SLOAN_rprime.csv'), 'angstroms'
    if input_str == 'u':
        return os.path.join(filter_folder, 'filters/SLOAN_u.csv'), 'angstroms'
    if input_str == 'u_prime':
        return os.path.join(filter_folder, 'filters/SLOAN_uprime.csv'), 'angstroms'
    if input_str == 'z':
        return os.path.join(filter_folder, 'filters/SLOAN_z.csv'), 'angstroms'
    if input_str == 'z_prime':
        return os.path.join(filter_folder, 'filters/SLOAN_zprime.csv'), 'angstroms'

    # Kepler and TESS
    if input_str.lower() == 'kepler':
        return os.path.join(filter_folder, 'filters/Kepler.csv'), 'angstroms'
    if input_str.lower() == 'tess':
        return os.path.join(filter_folder, 'filters/TESS.csv'), 'angstroms'


    raise ValueError(f'Unable to convert "{input_str}" into a path to a filter file')


#############################################################
#                         OUTPUTS                           #
#############################################################
def print_results(results, priorinfo, n_dof):
    '''
    Prints the results nicely to terminal

    Parameters
    ----------
    results : dynesty.results.Results
        The Dynesty results object, but must also have weights, cov and
        uncertainties as entries.
    priorinfo : transitfit.priorinfo.PriorInfo
        The PriorInfo object
    '''
    best = results.samples[np.argmax(results.logl)]

    print('\nBest fit results:')

    # We need to print out the results. Loop over each fitted
    for i, param_info in enumerate(priorinfo.fitting_params):
        param, tidx, fidx, eidx = param_info
        if not (param in priorinfo.limb_dark_coeffs and priorinfo.ld_fit_method == 'single'):
            if param in priorinfo.limb_dark_coeffs:
                # We need to convert the LDCs
                LDC_index = int(param[-1])
                LDC_vals = best[i - LDC_index : i + len(priorinfo.limb_dark_coeffs) - LDC_index]
                LDC_uncs = results.uncertainties[i - LDC_index : i + len(priorinfo.limb_dark_coeffs) - LDC_index]
                value = LDC_vals[LDC_index]
                unc = LDC_uncs[LDC_index]

            else:
                value = round(best[i], 6)
                unc = round(results.uncertainties[i], 6)

            if param in ['rp']:
                param = param +'_{}:\t'.format(int(fidx))
            #elif param in ['t0']:
            #    param = param +'_{}'.format(int(priorinfo._epoch_idx[i]))
            elif (param in priorinfo.detrending_coeffs + ['norm']) or (param in['t0'] and priorinfo.allow_ttv):
                param = param + '_t{}_f{}_e{}:'.format(int(tidx),int(fidx), int(eidx))
            elif param in priorinfo.limb_dark_coeffs and priorinfo.ld_fit_method in ['independent', 'coupled']:
                # All the LD coeffs are fitted separately and will write out
                param = param +'_{}:\t'.format(int(fidx))
            else:
                param += ':\t'
            print('{}\t {} ± {}'.format(param, value, unc))

    # Deal with 'single' fitting LD param mode
    if priorinfo.ld_fit_method == 'single':
        # We've only fitted one wavelengths' LD coeffs:
        # Estimate the rest of the limb darkening values and write.

        # Get the fitted values in a usable format
        best_single_ld = np.zeros(len(priorinfo.limb_dark_coeffs))
        single_uncertainty = np.zeros(len(priorinfo.limb_dark_coeffs))

        for i, ui in enumerate(priorinfo.limb_dark_coeffs):
            x = np.where(np.array(priorinfo.fitting_params) == ui)[0]
            best_single_ld[i] = best[x]
            single_uncertainty = results.uncertainties[x]

        # Now do the estimation
        best_ld_params = priorinfo.ld_handler.ldtk_estimate(best_single_ld)

        # We estimate the errors by using the ratios
        estim_errors = priorinfo.ld_handler.ldtk_estimate(single_uncertainty)

        # Loop through each filter and write out the values
        for i, fi in enumerate(best_ld_params):
            for j, qj in enumerate(priorinfo.limb_dark_coeffs):
                param = qj + '_{}:\t'.format(i)
                value = round(fi[j], 6)
                unc = round(estim_errors[i, j], 6)

                print('{}\t {} ± {}'.format(param, value, unc))

    best_chi2 = - results.logl.max()

    print('chi2:\t\t {}'.format(round(best_chi2, 5)))
    print('red chi2:\t {}'.format(round(best_chi2/n_dof, 5)))

def save_final_light_curves(lightcurves, priorinfo, results,
                            folder='./final_light_curves', folded=False):
    '''
    Applies detrending and normalisation to each light curve and saves to .csv

    Parameters
    ----------
    lightcurves : array_like, shape (n_telescopes, n_filters, n_epochs)
        An array of LightCurves. If no data exists for a point in the array
        then the entry should be `None`.
    priorinfo : transitfit.priorinfo.PriorInfo
        The PriorInfo object
    results : dynesty.results.Results
        The Dynesty results object, but must also have weights, cov and
        uncertainties as entries.
    folder : str, optional
        The folder to save the files to. Default is './final_light_curves'
    folded : bool, optional
        If True, will assume that the lightcurves provided are folded and
        will change the filenames accordingly. Default is False
    '''
    # Get the best values
    best = results.samples[np.argmax(results.logl)]
    best_dict = priorinfo._interpret_param_array(best)

    # Get the array of detrending coeffs:
    if priorinfo.detrend:
        # We need to combine the detrending coeff arrays into one
        # Each entry should be a list containing all the detrending
        # coefficients to trial.
        d = np.full(lightcurves.shape, None, object)

        for i in np.ndindex(d.shape):
            for coeff in np.ravel(priorinfo.detrending_coeffs):
                if best_dict[coeff][i] is not None:
                    if d[i] is None:
                        d[i] = [best_dict[coeff][i]]
                    else:
                        d[i].append(best_dict[coeff][i])

    os.makedirs(folder, exist_ok=True)

    # Loop over each light curve and apply detrending and normalisation
    for i in np.ndindex(lightcurves.shape):
        if lightcurves[i] is not None:

            telescope_idx = lightcurves[i].telescope_idx
            filter_idx = lightcurves[i].filter_idx
            epoch_idx = lightcurves[i].epoch_idx

            # Calculate the detrended light curve
            norm = best_dict['norm'][i]

            if priorinfo.detrend:
                detrended_flux, detrended_errors = lightcurves[i].detrend_flux(d[i], norm)
            else:
                detrended_flux, detrended_errors = lightcurves[i].detrend_flux(None, norm)

            # Calculate the value of the best fit light curve at the same times
            # First we set up the parameters
            params = batman.TransitParams()
            params.t0 = best_dict['t0'][i]
            params.per = best_dict['P'][i]
            params.rp = best_dict['rp'][i]
            params.a = best_dict['a'][i]
            params.inc = best_dict['inc'][i]
            params.ecc = best_dict['ecc'][i]
            params.w = best_dict['w'][i]
            params.limb_dark = priorinfo.limb_dark

            if priorinfo.fit_ld:
                best_q = np.array([best_dict[key][i] for key in priorinfo.limb_dark_coeffs])
            else:
                q = np.array([priorinfo.priors[key][i] for key in priorinfo.limb_dark_coeffs])
                for j in np.ndindex(q.shape):
                    q[j] = q[j]
                best_q = q

            # Convert from q to u
            params.u = priorinfo.ld_handler.convert_qtou(*best_q)

            m_sample_times = batman.TransitModel(params, lightcurves[i].times)
            time_wise_best_curve = m_sample_times.light_curve(params)

            # Calculate phase for each point in the curve
            phase = lightcurves[i].get_phases(best_dict['t0'][i], best_dict['P'][i])

            write_dict = []
            for j, tj in enumerate(lightcurves[i].times):
                write_dict.append({'Time' : tj,
                                   'Phase' : phase[j],
                                   'Normalised flux' : detrended_flux[j],
                                   'Uncertainty' : detrended_errors[j],
                                   'Best fit curve' : time_wise_best_curve[j]})

            if folded:
                fname = 'filter_{}_FOLDED.csv'.format(filter_idx)
            else:
                fname = 't{}_f{}_e{}_detrended.csv'.format(telescope_idx, filter_idx, epoch_idx)

            with open(os.path.join(folder, fname), 'w') as f:
                columns = ['Time', 'Phase', 'Normalised flux', 'Uncertainty', 'Best fit curve']
                writer = csv.DictWriter(f, columns)
                writer.writeheader()
                writer.writerows(write_dict)



'''
Class to handle limb darkening parameters through PyLDTK

'''

import numpy as np
from ldtk import LDPSetCreator, BoxcarFilter, TabulatedFilter
import os
from collections.abc import Iterable

_implemented_ld_models = ['linear', 'quadratic', 'nonlinear', 'power2', 'squareroot']

class LDTKHandler:
    '''
    The LDTKHandler provides an easy way to interface ldtk with TransitFit.

    Parameters
    ----------
    host_T : tuple
        The effective temperature of the host star, in Kelvin, given as a
        (value, uncertainty) pair.
    host_logg : tuple
        The log_10 of the surface gravity of the host star, with gravity
        measured in cm/s2. Should be given as a (value, uncertainty) pair.
    host_z : tuple
        The metalicity of the host, given as a (value, uncertainty) pair.
    filters : array_like
        The set of filters, given in [low, high] limits for the wavelengths
        with the wavelengths given in nanometers if a uniform filter is to
        be used, or [[wavelength...], [transmission]] if a fully-defined
        profile is being used. The ordering of the filters should
        correspond to the filter_idx parameter used elsewhere.
    ld_method : str, optional
        The model of limb darkening to use. Allowed values are 'linear',
        'quadratic', 'squareroot', 'power2', and 'nonlinear'. Default is
        'quadratic'.
    n_samples : int, optional
        The number of limb darkening profiles to create. Passed to
        ldtk.LDPSetCreator.create_profiles(). Default is 20000.
    do_mc : bool, optional
        If True, will use MCMC to estimate coefficient uncertainties more
        accurately. Default is False.
    cache_path : str, optional
        This is the path to cache LDTK files to. If not specified, will
        default to the LDTK default
    ldtk_uncertainty_multiplier: float, optional
        (From LDTK:) The uncertainty multiplier ϵ is a subjective factor
        that defines how strongly the LD profile (or the prior created
        from it) constrains the final analysis (that is, how much we
        trust the stellar atmosphere models used to create the profiles.)
    '''
    def __init__(self, host_T, host_logg, host_z, filters,
                 ld_model='quadratic', n_samples=20000, do_mc=False,
                 cache_path=None, ldtk_uncertainty_multiplier=1.):

        # Sanity checks
        if not ld_model in _implemented_ld_models:
            raise ValueError('Unrecognised ld_model {}'.format(ld_model))

        self.default_model = ld_model

        # Set up the filters
        #print('Setting up filters')
        ldtk_filters = []
        for i, f in enumerate(filters):
            if isinstance(f[0], Iterable):
                # We have been passed a full filter profile, set up
                # TabulatedFilter
                # Work out if the profile is in percent or fraction - is
                # anything bigget than 1?
                if np.any(f[1] > 1):
                    tmf = 1e-2
                else:
                    tmf = 1
                ldtk_filters.append(TabulatedFilter(i, f[0], f[1], tmf))
            else:
                ldtk_filters.append(BoxcarFilter(i, f[0], f[1]))

        # Make the set creator, downloading data files if required
        if cache_path is not None:
            os.makedirs(cache_path, exist_ok=True)
        #print('Making LD parameter set creator.')
        #print('This may take some time as we may need to download files...')
        set_creator = LDPSetCreator(teff=host_T, logg=host_logg, z=host_z,
                                    filters=ldtk_filters, cache=cache_path,
                                    dataset='visir-lowres')

        # Get the LD profiles from the set creator
        #print('Obtaining LD profiles')
        self.profile_set = set_creator.create_profiles(nsamples=n_samples)
        self.profile_set.set_uncertainty_multiplier(ldtk_uncertainty_multiplier)

        # Find the 'best values' for each filter and then find the ratios
        # compared to the first.
        #print('Finding coefficients and ratios')
        self.coeffs = {}
        self.ratios = {}

        self._power2_available = True

        for model in _implemented_ld_models:
            try:
                self.coeffs[model] = self._extract_best_coeffs(model)
                self.ratios[model] = self.coeffs[model][0] / self.coeffs[model][0][0]
            except Exception as e:
                print(e)
                print(f'Unable to initialise {model} model')
                self._power2_available = False

    def estimate_values(self, ld0_values, ld_model):
        '''
        If given a set of LD param values for filter 0, will estimate the LD
        parameters for all filters based on the ratios between the best values
        found in initialisation.

        Parameters
        ----------
        ld1_values : float or array_like
            The LD parameters for filter 0
        ld_model : str
            The limb darkening model to use

        Returns
        -------
        all_ld_values : array_like, shape (n_filters, n_coeffs)
            The estimated limb darkening parameters
        '''
        if ld_model == 'power2' and not self._power2_available:
            raise ValueError('power2 model is not available. If you want to use this, please use the development version of ldtk available on https://github.com/hpparvi/ldtk, rather than the pypi version.')

        return ld0_values * self.ratios[ld_model]

    def _extract_best_coeffs(self, ld_model, do_mc=False):
        '''
        Extracts the best values for a limb darkening model for the filters

        Parameters
        ----------
        ld_model : str
            The limb darkening model to obtain the values for.

        Returns
        -------
        coeffs : array_like, shape (n_filters, n_coeffs)
            The coefficients for each filter
        err : array_like, shape (n_filters, n_coeffs)
            The uncertainty on each of the coefficients
        '''
        if ld_model == 'linear':
            coeff, err = self.profile_set.coeffs_ln(do_mc=do_mc)
        elif ld_model == 'quadratic':
            coeff, err = self.profile_set.coeffs_qd(do_mc=do_mc)
        elif ld_model == 'nonlinear':
            coeff, err = self.profile_set.coeffs_nl(do_mc=do_mc)
        elif ld_model == 'power2':
            if not self._power2_available:
                raise ValueError('power2 model is not available. If you want to use this, please use the development version of ldtk available on https://github.com/hpparvi/ldtk, rather than the pypi version.')
            coeff, err = self.profile_set.coeffs_p2(do_mc=do_mc)
        elif ld_model == 'squareroot':
            coeff, err = self.profile_set.coeffs_sq(do_mc=do_mc)

        else:
            raise ValueError('Unrecognised ld_model {}'.format(ld_model))

        return coeff, err

    def lnlike(self, coeffs, ld_model):
        '''
        Evaluates the log likelihood for a set of coefficients

        Parameters
        ----------
        coeffs : array_like, shape (n_filters, n_coeffs)
            The coefficients to evaluate the log likelihood for.
        ld_model : str, optional
            The model to use. Defaults to self.default_model

        '''
        if ld_model == 'linear':
            return self.profile_set.lnlike_ln(coeffs)
        if ld_model == 'quadratic':
            return self.profile_set.lnlike_qd(coeffs)
        if ld_model == 'nonlinear':
            return self.profile_set.lnlike_nl(coeffs)
        if ld_model == 'power2':
            if not self._power2_available:
                raise ValueError('power2 model is not available. If you want to use this, please use the development version of ldtk available on https://github.com/hpparvi/ldtk, rather than the pypi version.')
            return self.profile_set.lnlike_p2(coeffs)
        if ld_model == 'squareroot':
            return self.profile_set.lnlike_sq(coeffs)

        raise ValueError('Unrecognised ld_model {}'.format(ld_model))



'''
A function which will run everything when given a path to light curve and
priors!
'''

from .io import read_priors_file, read_input_file, read_filter_info, parse_data_path_list, read_data_path_array, parse_priors_list, parse_filter_list
from .retriever import Retriever
from ._utils import check_batches

import numpy as np
import os


def run_retrieval(data_files, priors, filter_info=None,
                  detrending_list=[['nth order', 1]],
                  limb_darkening_model='quadratic',
                  ld_fit_method='independent', fitting_mode='auto',
                  max_batch_parameters=25, batch_overlap=2,
                  host_T=None, host_logg=None, host_z=None, host_r=None,
                  nlive=300, dlogz=None, maxiter=None, maxcall=None,
                  dynesty_sample='rslice', dynesty_bounding='multi',
                  normalise=True, detrend=True,
                  results_output_folder='./output_parameters',
                  final_lightcurve_folder='./fitted_lightcurves',
                  summary_file='summary_output.csv',
                  full_output_file='full_output.csv',
                  plot_folder='./plots', plot=True,
                  marker_color='dimgrey', line_color='black', ldtk_cache=None,
                  ldtk_samples=20000, do_ld_mc=False, data_skiprows=0,
                  allow_ttv=False, filter_delimiter=None,
                  detrending_limits=None, normalise_limits=None, bin_data=False,
                  cadence=2, binned_color='red', walks=100, slices=10, 
                  n_procs=1, check_batchsizes=False, median_normalisation=False,
                  error_scaling=False, error_scaling_limits=None, 
                  ldtk_uncertainty_multiplier=1.):
    '''
    Runs a full retrieval of posteriors using nested sampling on a transit
    light curve or a set of transit light curves. For more guidance on the use
    of input files and structuring, see the TransitFit documentation.

    Parameters
    ----------
    data_files : str
        The path to the data input .csv file, which contains the paths to the
        light curve data.

    priors : str
        Path to a .csv file containing prior information on each variable to be
        fitted.

    filter_info : str, optional
        Path to a .csv file containing information on the wavelengths of the
        filters that observations were made at.

        This is required if ld_fit_method is `'single'` or `'coupled'`. If not
        None and host_T, host_logg and host_z are not None, retrieval will
        include fitting realistic limb darkening parameters for the filters.
        Default is None.

    detrending_list : array_like, shape (n_detrending_models, 2)
        A list of different detrending models. Each entry should consist
        of a method and a second parameter dependent on the method.
        Accepted methods are
            ``['nth order', order]``

            ``['custom', function, [global fit indices], [filter fit indices], [epoch fit indices]]``

            ``['off', ]``
        ``function`` here is a custom detrending function. TransitFit assumes
        that the first argument to this function is times and that all
        other arguments are single-valued - TransitFit cannot fit
        list/array variables. If 'off' is used, no detrending will be
        applied to a light curve using this model.

        If a custom function is used, and some inputs to the function
        should not be fitted individually for each light curve, but should
        instead be shared either globally, within a given filter, or within
        a given epoch, the indices of where these fall within the arguments
        of the detrending function should be given as a list. If there are
        no indices to be given, then use an empty list: []
        e.g. if the detrending function is given by::

            foo(times, a, b, c, t0, P):
                # do something

        and a should be fitted globally, then the entry in the method_list
        would be ``['custom', foo, [1], [], []]``.

    limb_darkening_model : str, optional
        The limb darkening model to use. Allowed models are

            - ``'linear'``
            - ``'quadratic'``
            - ``'squareroot'``
            - ``'power2'``
            - ``'nonlinear'``

        With the exception of the non-linear model, all models are constrained
        by the method in Kipping (2013), which can be found at
        https://arxiv.org/abs/1308.0009.
        Default is 'quadratic'.

    ld_fit_method : str, optional
        Determines the mode of fitting of limb darkening parameters. The
        available modes are:
            - ``'coupled'`` : all limb darkening parameters are fitted
              independently, but are coupled to a wavelength dependent
              model based on the host parameters through `ldkt`
            - ``'single'`` : LD parameters are still tied to a model, but only
              the first filter is actively fitted. The remaining filters
              are estimated based off the ratios given by ldtk for a host
              with the given parameters. This mode is useful for a large
              number of filters, as 'coupled' or 'independent' fitting will
              lead to much higher computation times.
            - ``'independent'`` : Each LD coefficient is fitted separately for
              each filter, with no coupling to the ldtk models.
            - ``'off'`` : no limb darkening fitting will occurr. If using this
              mode, it it strongly recommended to set values for the
              Kipping q parameters using the priors file.
        Default is `'independent'`

    fitting_mode : {``'auto'``, ``'all'``, ``'2_stage'``, ``'folded'``, ``'batched'``}, optional
        The approach TransitFit takes towards limiting the number of parameters
        being simultaneously fitted. The available modes are:
        - ``'auto'`` : Will calculate the number of parameters required to fit
          all the data simulataneously. If this is less than max_parameters,
          will set to ``'all'`` mode, else will set to ``'folded'`` if at least
          one filter has at least 3 epochs in it. Otherwise will set to
          ``'batched'``
        - ``'all'`` : Fits all parameters simultaneously, with no folding or
          batching of curves. Should be used with caution when fitting very
          large (~< 30) numbers of parameters.
        - ``'2_stage'`` : Fits in 2 stages, first detrending the light curves
          and then fitting the detrended curves simultaneously, using the
          ``'batched'`` approach if required.
        - ``'folded'`` : Useful for fitting curves with multiple epochs for each
          filter. TransitFit will fit each filter separately and produce a
          period-folded light curve for each filter, before fitting these
          simultaneously, using the ``'batched'`` approach if required.
        - ``'batched'`` : Useful for large numbers of light curves with
          relatively few shared filters, so ``'folded'`` loses large amounts of
          multi-epoch information. This mode splits the filters into sets of
          overlapping batches, runs each batch and uses the weighted means of
          each batch to produce a final result.
        Default is ``'auto'``.

    max_batch_parameters : int, optional
        The maximum number of parameters to use in a single retrieval when
        using ``'folded'`` or ``'batched'`` fitting modes. Default is 25.

    batch_overlap : in, optional
        The number of filters or epochs to overlap adjacent batches by where
        possible. This ensures that adjacent batches share information. Default
        is 2.

    host_T : tuple or None, optional
        The effective temperature of the host star, in Kelvin. Should be given
        as a (value, uncertainty) pair. Required if ld_fit_method is
        ``'single'`` or ``'coupled'``. Default is None.

    host_logg : tuple or None, optional
        The log_10 of the surface gravity of the host star, with gravity
        measured in cm/s2. Should be given as a (value, uncertainty) pair.
        Required if ld_fit_method is ``'single'`` or ``'coupled'``. Default is
        None

    host_z : tuple or None, optional
        The metalicity of the host, given as a (value, uncertainty) pair.
        Required if ld_fit_method is ``'single'`` or ``'coupled'``. Default is
        None

    host_r : tuple or None, optional
        The host radius in Solar radii, given as a (value, uncertainty) pair.
        Required for conversion of host-planet separation from AU to host radii

    nlive : int, optional
        The number of live points to use in the nested sampling retrieval.

    normalise : bool, optional
        If True, will assume that the light curves have not been normalised and
        will fit normalisation constants within the retrieval. The range to
        fit normalisation constants c_n are automatically detected using
            ``1/f_max <= c_n <= 1/f_min``
        as the default range, where ``f_min`` and ``f_max`` are the minimum and
        maximum flux values for a given light curve. Default is ``True``.

    detrend : bool, optional
        If False, no detrending will be attempted, even if specified by
        detrending list. Default is ``True``

    dlogz : float, optional
        Retrieval iteration will stop when the estimated contribution of the
        remaining prior volume to the total evidence falls below this
        threshold. Explicitly, the stopping criterion is
        ``ln(z + z_est) - ln(z) < dlogz``, where ``z`` is the current evidence
        from all saved samples and z_est is the estimated contribution from
        the remaining volume. The default is ``1e-3 * (nlive - 1) + 0.01``.

    maxiter : int or None, optional
        The maximum number of iterations to run. If None, will
        continue until stopping criterion is reached. Default is ``None``.

    maxcall : int or None, optional
        The maximum number of likelihood calls in retrieval. If None, will
        continue until stopping criterion is reached. Default is ``None``.

    dynesty_sample : str, optional
        Method used to sample uniformly within the likelihood constraint,
        conditioned on the provided bounds. Unique methods available are:
        uniform sampling within the bounds(``'unif'``), random walks with fixed
        proposals (``'rwalk'``), random walks with variable (“staggering”)
        proposals (``'rstagger'``), multivariate slice sampling along preferred
        orientations (``'slice'``), “random” slice sampling along all
        orientations (``'rslice'``), “Hamiltonian” slices along random
        trajectories (``'hslice'``), and any callable function which follows
        the pattern of the sample methods defined in dynesty.sampling.
        'auto' selects the sampling method based on the dimensionality of
        the problem (from ``ndim``). When ndim < 10, this defaults to ``'unif'``.
        When 10 <= ``ndim`` <= 20, this defaults to ``'rwalk'``. When ndim > 20,
        this defaults to ``'hslice'`` if a gradient is provided and ``'slice'``
        otherwise. ``'rstagger'`` and ``'rslice'`` are provided as alternatives for
        ``'rwalk'`` and ``'slice'``, respectively. Default is ``'rslice'``.

    dynesty_bounding : {``'none'``, ``'single'``, ``'multi'``, ``'balls'``, ``'cubes'``}, optional
        The decomposition to use in sampling. Default is 'multi'

    results_output_folder : str, optional
        Folder to save results to. TransitFit will create subfolders within
        this if folded or batched runs are being used. Default is
        ``'./output_parameters'``

    fitted_lightcurve_folder : str, optional
        The folder to save fitted light curves to. These files contain the
        normalised and detrended light curves, as well as the best fit curve.
        Default is ``'./fitted_lightcurves'``

    summary_file : str, optional
        The file to save the summarised final parameter results to. These are
        calculated by taking weighted averages over any batched fitting.
        Default is ``'summary_output.csv'``

    full_output_file : str, optional
        The file to save the full, non-summarised results to. This file gives
        the results for each batch, without averaging over batches to get
        summaried results. Default is ``'full_output.csv'``

    plot_folder : str, optional
        Path to folder to save plots to. Default is ``'./plots'``

    plot : bool, optional
        If ``True``, will plot all fitted light curves within the fitting routine,
        including any from partial fitting (eg, single filter modes). Default
        is ``True``.

    marker_color : matplotlib color, optional
        The colour to plot data points on plots. Default is ``'dimgray'``.

    line_colour : matplotlib color, optional
        The colour to plot best fit light curves on plots. Default is ``'black'``.

    ldtk_cache : str, optional
        This is the path to cache LDTK files to. If not specified, will
        default to the LDTK default.

    ldtk_samples : int, optional
        Controls the number of samples taken by PyLDTk when calculating LDCs
        when using ``'coupled'`` or ``'single'`` modes for limb darkening
        fitting. Default is ``20000``

    do_ld_mc : bool, optional
        If ``True``, will use MCMC sampling to more accurately estimate the
        uncertainty on intial limb darkening parameters provided by PyLDTk.
        Default is ``False``.

    data_skiprows : int, optional
        The number of rows to skip when reading in light curve data from a .txt
        file. Default is ``0``.

    allow_ttv : bool, optional
        If ``True``, will fit t0 for each epoch individually. Default is
        ``False``.

    filter_delimiter : str, optional
        The delimiter in filter profile files. Default is ``None``, which will
        lead to ``pandas`` trying to auto detect the delimiter.

    detrending_limits : list, optional
        The bounds on detrending coefficients, given as (lower, upper) pair for
        each detrending method. IF not provided, will default to ±10

    bin_data : bool, optional
        If True, any folded light curves will be plotted with data binned to an
        observing cadence given by `cadence`. Default is False.

    cadence : float, optional
        The observing cadence, in minutes, to bin data to if `bin_data` is
        True. Default is 2 (mirroring TESS observations)

    binned_color : str, optional
        The color to use for binned data. Default is `'red'`.

    n_procs : int, optional
        The maximum number of processes to use when running batches. If >1,
        will run batches in parallel. Default is 1.
    
    median_normalisation : bool, optional
        Normalising lightcurves by the median of the flux value reduces the runtime in many cases.
        Default is False.

    error_scaling : bool, optional
            If True, scales the errorbars in the lightcurves following 
            https://emcee.readthedocs.io/en/stable/tutorials/line/
    
    error_scaling_limits: list, optional
            If error_scaling=True, this is the limit of the parameter.
    
    ldtk_uncertainty_multiplier: float, optional
        (From LDTK:) The uncertainty multiplier ϵ is a subjective factor
        that defines how strongly the LD profile (or the prior created
        from it) constrains the final analysis (that is, how much we
        trust the stellar atmosphere models used to create the profiles.)

    Returns
    -------
    results : dict
        The results returned by ``Retriever.run_dynesty()``
    '''
    print('Starting transitfit')

    # If required to check for uneven batch sizes
    if check_batchsizes:
        data_files = check_batches(allow_ttv, data_files)
    # Load in the data and work out number of telescopes, filters, and epochs
    lightcurves, detrending_index_array = read_input_file(data_files)

    n_telescopes = lightcurves.shape[0]
    n_filters = lightcurves.shape[1]
    n_epochs = lightcurves.shape[2]

    # Set up the Retriever
    retriever = Retriever(data_files, priors, n_telescopes, n_filters, n_epochs,
                          filter_info, detrending_list, limb_darkening_model,
                          host_T, host_logg, host_z, host_r, ldtk_cache,
                          ldtk_samples, do_ld_mc, data_skiprows, allow_ttv,
                          filter_delimiter, detrending_limits, normalise, 
                          normalise_limits,detrend,median_normalisation,
                          error_scaling, error_scaling_limits, 
                          ldtk_uncertainty_multiplier)

    # Run the retrieval!
    results = retriever.run_retrieval(ld_fit_method, fitting_mode,
                                      max_batch_parameters, maxiter, maxcall,
                                      dynesty_sample, nlive, dlogz, plot,
                                      results_output_folder,
                                      final_lightcurve_folder, summary_file,
                                      full_output_file, plot_folder,
                                      marker_color, line_color, dynesty_bounding, normalise,
                                      detrend, batch_overlap, bin_data, cadence, binned_color,
                                      walks, slices, n_procs)

    return results



'''
detrender.py

A module to deal with arbitrary detrending

'''

import numpy as np
from inspect import signature
import inspect

class DetrendingFunction:
    '''
    The DetrendingFunction is designed to handle detrending with an arbitrary
    function.

    Basically gets around dealing with varying number of args in a neat way

    Parameters
    ----------
    function : function
        The detrending function. Must have signature of f(times, a, b, c,...,t0,P)
        where a, b, c etc are the detrending coefficients, t0 is the time of 
        conjunction and P is period.
    '''
    def __init__(self, function,method=None):
          

        # Store the function
        self.function = function

        # Do some analysis of the signature to get info on arg, kwargs and the like
        sig = signature(self.function)
        params = sig.parameters

        args = 0
        kwargs = 0
        for p in params:
            if params[p].default == inspect._empty:
                # No default value given - must be supplied
                args += 1
            else:
                kwargs += 1
        if method == 'custom':
            args=args-2 
            # For custom functions, 2 arguments are subtracted because 
            # they correpond to 't0' and 'P'

        self.n_required_args = args
        self.n_kwargs = kwargs
        
        # -1 assumes that the first entry is lightcurve, so won't be fitted
        try:
            self.n_params = function.order
        except:
            self.n_params = len(params)

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)



'''
Some detrending functions for use with LightCurve
'''

import numpy as np


class NthOrderDetrendingFunction:
    '''
    Arbitrary order detrending function which conserves flux.

    Parameters
    ----------
    order : int
        The detrending function order. Must be greater than 0.
    '''
    def __init__(self, order):

        order = int(order)
        if not order > 0:
            raise ValueError('Order must be greater than 0')

        self.order = order


    def __call__(self, lightcurve, *args):
        '''
        Detrends the lightcurve and returns the flux values
        '''
        
        if not int(len(args)-2) == int(self.order):
            
            raise ValueError('Number of arguments {} supplied does not match the order {} of the detrending function.'.format(len(args)-2, self.order))

        
        t0=args[-2]
        p=args[-1]
        # We calculate t01 which is time of conjucntion for the particular lightcurve. 
        t01=np.max(lightcurve.times)-((np.max(lightcurve.times)-t0)%p)
        #times = lightcurve.times-np.mean(lightcurve.times)
        times = lightcurve.times-t01

        #times = lightcurve.times - np.floor((np.min(lightcurve.times)))
        #self.time0= time0
        
        vals = np.zeros(len(times))
        for i in range(0, self.order):
            #vals += args[i] * (times ** (i+1) - np.mean(times ** (i+1))) # Original function in JH version
            vals += args[i] * ((times)** (i+1))
            #vals += args[i] * ((times)** (i+1))

        return lightcurve.flux - vals
        """
        '''
        Detrends the lightcurve and returns the flux values
        '''
        if not len(args) == self.order:
            raise ValueError('Number of arguments {} supplied does not match the order {} of the detrending function.'.format(len(args), self.order))

        # We rescale the times because in BJD the numbers can get huge!
        times = lightcurve.times - np.floor((np.min(lightcurve.times)))

        #self.time0= time0

        vals = np.zeros(len(times))
        for i in range(0, self.order):
            #vals += args[i] * (times ** (i+1) - np.mean(times ** (i+1))) # Original function in JH version
            vals += args[i] * ((times - np.mean(times))** (i+1)) # We need to replace times with times0

        return lightcurve.flux - vals"""



'''
setup script for transitfit

'''

import setuptools
import os
import codecs

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            # __version__ = "0.9"
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="transitfit",
    version=get_version('transitfit/__init__.py'),
    author="Joshua Hayes and collaborators",
    author_email="Eamonn.Kerins@manchester.ac.uk",
    description="A package to fit exoplanet transit light curves",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SPEARNET/TransitFit",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta"
    ],
    python_requires='>=3.6',
    install_requires=['numpy','batman-package', 'dynesty',  'matplotlib',
                      'pandas', 'ldtk>=1.5.0', 'corner', 
                      'semantic_version', 'scipy', 'statsmodels'],
    zip_safe=False,
    package_data={'transitfit' : ['../filters/*.csv']}
)



"""
Object which actually handles the retrieval process
"""

import numpy as np
from dynesty import NestedSampler
from copy import deepcopy
import os, csv
import traceback
import multiprocessing as mp
import pandas as pd
from collections.abc import Iterable

import io


from .plotting import plot_individual_lightcurves, quick_plot
from .detrending_funcs import NthOrderDetrendingFunction
from .detrender import DetrendingFunction
from .io import (
    read_input_file,
    read_priors_file,
    parse_priors_list,
    read_filter_info,
    parse_filter_list,
    print_results,
)
from ._likelihood import LikelihoodCalculator
from ._utils import  weighted_avg_and_std

# Parameters and if they are global, filter-specific or lightcurve-specific
global_params = ["P", "ecc", "a", "inc", "w"]

filter_dependent_params = ["rp", "q0", "q1", "q2", "q3", "u0", "u1", "u2", "u3"]

lightcurve_dependent_params = [
    "norm",
    "d0",
    "d1",
    "d2",
    "d3",
    "d4",
    "d5",
    "d6",
    "d7",
    "d8",
]

#_default_detrending_limits = (-10,10)
_default_detrending_limits = (-0.1,0.1)

from .output_handler import OutputHandler, Results, ResultsException


class Retriever:
    """
    Handles the nested sampling retrieval and acts as a high-level interface

    Parameters
    ----------
    data_files : str
        Path to the data input config file
    priors : str
        Path to the priors config file
    n_telescopes : int
        The number of different telescopes used in the data set
    n_filters : int
        The number of different filters used in the data set
    n_epochs : int
        The number of different epochs used in the data set
    filter_info : str, optional
        The path to the filter profiles config file
    detrending_list: list, optional
        The different detrending models to use
    limb_darkening_model : str
        The limb darkening model to use
    host_T : tuple or None, optional
        The effective temperature of the host star, in Kelvin. Should be given
        as a (value, uncertainty) pair. Required if ld_fit_method is
        ``'single'`` or ``'coupled'``. Default is None.
    host_logg : tuple or None, optional
        The log_10 of the surface gravity of the host star, with gravity
        measured in cm/s2. Should be given as a (value, uncertainty) pair.
        Required if ld_fit_method is ``'single'`` or ``'coupled'``. Default is
        None
    host_z : tuple or None, optional
        The metalicity of the host, given as a (value, uncertainty) pair.
        Required if ld_fit_method is ``'single'`` or ``'coupled'``. Default is
        None
    host_r : tuple or None, optional
        The host radius in Solar radii, given as a (value, uncertainty) pair.
        Required for conversion of host-planet separation from AU to host radii
    ldtk_cache : str, optional
        This is the path to cache LDTK files to. If not specified, will
        default to the LDTK default.
    n_ld_samples : int, optional
        Controls the number of samples taken by PyLDTk when calculating LDCs
        when using ``'coupled'`` or ``'single'`` modes for limb darkening
        fitting. Default is ``20000``
    do_ld_mc : bool, optional
        If ``True``, will use MCMC sampling to more accurately estimate the
        uncertainty on intial limb darkening parameters provided by PyLDTk.
        Default is ``False``.
    data_skiprows : int, optional
        The number of rows to skip when reading in light curve data from a .txt
        file. Default is ``0``.
    allow_ttv : bool, optional
        If ``True``, will fit t0 for each epoch individually. Default is
        ``False``.
    filter_delimiter : str, optional
        The delimiter in filter profile files. Default is ``None``, which will
        lead to ``pandas`` trying to auto detect the delimiter.
    detrending_limits : list, optional
        The bounds on detrending coefficients, given as (lower, upper) pair for
        each detrending method. If not provided, will default to ±10
    error_scaling : bool, optional
            If True, scales the errorbars in the lightcurves following 
            https://emcee.readthedocs.io/en/stable/tutorials/line/
    error_scaling_limits: list, optional
            If error_scaling=True, this is the limit of the parameter.
    ldtk_uncertainty_multiplier: float, optional
            (From LDTK:) The uncertainty multiplier ϵ is a subjective factor
            that defines how strongly the LD profile (or the prior created
            from it) constrains the final analysis (that is, how much we
            trust the stellar atmosphere models used to create the profiles.)
    """

    def __init__(
        self,
        data_files,
        priors,
        n_telescopes,
        n_filters,
        n_epochs,
        filter_info=None,
        detrending_list=[["nth order", 1]],
        limb_darkening_model="quadratic",
        host_T=None,
        host_logg=None,
        host_z=None,
        host_r=None,
        ldtk_cache=None,
        n_ld_samples=20000,
        do_ld_mc=False,
        data_skiprows=0,
        allow_ttv=False,
        filter_delimiter=None,
        detrending_limits=None,
        normalise=True,
        normalise_limits=None,
        detrend=True,
        median_normalisation=False,
        error_scaling=False,
        error_scaling_limits=None,
        ldtk_uncertainty_multiplier=1.,
    ):

        ###################
        # Save input data #
        ###################
        self._data_input = data_files
        self._prior_input = priors
        self._filter_input = filter_info

        self.n_telescopes = n_telescopes
        self.n_filters = n_filters
        self.n_epochs = n_epochs

        self.allow_ttv = allow_ttv

        self.detrending_info = detrending_list

        # self.detrending_limits = detrending_limits
        if detrending_limits is None:
            self.detrending_limits = np.array(
                [_default_detrending_limits for i in range(len(self.detrending_info))]
            )

        else:
            """
            detrending_limits = np.array(detrending_limits)
            if not detrending_limits.ndim == 2:
                raise ValueError(f'Detrending limits should be provided as a list of length {len(self.detrending_info)} where each entry is the  [lower, upper] limits on each method.')
            if not detrending_limits.shape[1] == 2:
                raise ValueError(f'Detrending limits should be provided as a list of length {len(self.detrending_info)} where each entry is the  [lower, upper] limits on each method.')
            """
            self.detrending_limits = detrending_limits

        # Host info
        self.host_T = host_T
        self.host_logg = host_logg
        self.host_z = host_z
        self.host_r = host_r

        # Limb darkening things
        self.limb_darkening_model = limb_darkening_model
        self.n_ld_samples = n_ld_samples
        self.do_ld_mc = do_ld_mc
        self.ldtk_cache = ldtk_cache
        self.ldtk_uncertainty_multiplier = ldtk_uncertainty_multiplier

        ###########################################################################
        # Now read in things from files to get Filters, full priors, full lc data #
        ###########################################################################
        # Read in the filters
        if self._filter_input is None:
            self.filters = None
        elif type(self._filter_input) == str:
            self.filters = read_filter_info(self._filter_input, filter_delimiter)
        else:
            self.filters = parse_filter_list(self._filter_input, filter_delimiter)

        # Load in the full LightCurve data and detrending index array
        self.all_lightcurves, self.detrending_index_array = read_input_file(
            data_files, data_skiprows
        )

        if median_normalisation:
            print("Normalising lightcurves...")
            #normalise_limits=[0.95,1.05]
            for i in np.ndindex(self.all_lightcurves.shape):
                scale=np.median(self.all_lightcurves[i].flux)
                self.all_lightcurves[i].flux = self.all_lightcurves[i].flux/scale
                self.all_lightcurves[i].errors = self.all_lightcurves[i].errors/scale
        
        self.error_scaling = error_scaling
        if error_scaling:
            print('Scaling of errors initialised')
            if error_scaling_limits is None or len(error_scaling_limits)!=2:
                self.scaling_limits = [1e-7,10]
            else:
                if isinstance(error_scaling_limits[0], Iterable):
                    for esl in range(len(error_scaling_limits)):
                        error_scaling_limits[esl] = np.sort(np.abs(error_scaling_limits[esl]))
                        if error_scaling_limits[esl][0]<=0: 
                            error_scaling_limits[esl][0] = 1e-7
                    self.scaling_limits = error_scaling_limits

                else:
                    error_scaling_limits = np.sort(np.abs(error_scaling_limits))
                    if error_scaling_limits[0]<=0: 
                        error_scaling_limits[0] = 1e-7
                    self.scaling_limits = error_scaling_limits


        if detrend:
            # intialise detrending in each light curve
            for i, lc in np.ndenumerate(self.all_lightcurves):
                if lc is not None:
                    detrending_idx = self.detrending_index_array[i]
                    detrending_method = detrending_list[detrending_idx]

                    if detrending_method[0] == "nth order":
                        lc.set_detrending(
                            "nth order",
                            order=detrending_method[1],
                            method_idx=detrending_idx,
                        )
                    elif detrending_method[0] == "custom":
                        lc.set_detrending(
                            "custom",
                            function=detrending_method[1],
                            method_idx=detrending_idx,
                        )
                    elif detrending_method[0] == "off":
                        lc.set_detrending("off", method_idx=detrending_idx)
                    else:
                        raise ValueError(
                            f"Unrecognised detrending method {detrending_method[0]}"
                        )
        self.normalise_limits=normalise_limits
        # Initialse normalisation in each light curve
        if normalise:
            for i, lc in np.ndenumerate(self.all_lightcurves):
                if lc is not None:
                    lc.set_normalisation(self.normalise_limits)

        self.n_total_lightcurves = np.sum(self.all_lightcurves != None)

        # Get the FULL prior.
        # Assume independent ld method - it doesn't actually matter
        # since this never gets used.
        self._full_prior, _ = self._get_priors_and_curves(
            self.all_lightcurves, "independent", detrend=detrend, suppress_warnings=True, normalise=normalise
        )

        # From the full prior, get a bunch of info on numbers of parameters
        # Get the number of parameters in the limb darkening
        self.n_ld_params = len(self._full_prior.limb_dark_coeffs)
        self.ld_coeffs = self._full_prior.limb_dark_coeffs

        # Get the number of globally fitted parameters
        # Work out the possible global parameters based on allow_ttv
        global_params = ["P", "ecc", "a", "inc", "w"]
        if not self.allow_ttv:
            global_params.append("t0")
        self.n_global_params = 0
        for param in global_params:
            if param in np.array(self._full_prior.fitting_params)[:, 0]:
                self.n_global_params += 1
        
        if detrend:
            # Calculate the detrending models - used in calculating the number of
            # parameters used in a fitting batch
            self.detrending_models = []
            for mode in self.detrending_info:
                if mode[0] == "nth order":
                    function = NthOrderDetrendingFunction(mode[1])
                    self.detrending_models.append(DetrendingFunction(function))
                elif mode[0] == "custom":
                    function = mode[1]
                    self.detrending_models.append(DetrendingFunction(function,mode[0]))
                elif mode[0] == "off":
                    self.detrending_models.append(None)

        # Initialise the OutputWriter
        self.output_handler = OutputHandler(
            self.all_lightcurves, self._full_prior, self.host_r
        )


    ##########################################################
    #               RETRIEVAL RUNNING METHODS                #
    ##########################################################
    def _run_dynesty(
        self,
        lightcurves,
        priors,
        maxiter=None,
        maxcall=None,
        sample="auto",
        nlive=300,
        dlogz=None,
        bound="multi",
        plot_folder="./plots",
        walks=100,
        slices=10,
    ):
        """
        Runs dynesty on the given lightcurves with the given priors. Returns
        the result.

        Returns
        -------
        results
        ndof
        """
        # Get number of dimensions and degrees of freedom
        n_dims = len(priors.fitting_params)

        # Calculate the number of degrees of freedom - how many data points do we have?
        n_dof = 0
        for i in np.ndindex(lightcurves.shape):
            if lightcurves[i] is not None:
                n_dof += len(lightcurves[i].times)

        # Make a LikelihoodCalculator
        likelihood_calc = LikelihoodCalculator(lightcurves, priors)
        print(priors)

        #######################################################################
        #######################################################################
        # Now we define the prior transform and ln likelihood function for
        # dynesty to use
        def prior_transform(cube):
            return priors._convert_unit_cube(cube)

        def lnlike(cube):
            params = priors._interpret_param_array(cube)

            ln_likelihood= likelihood_calc.find_likelihood(params)

            if priors.fit_ld and not priors.ld_fit_method == "independent":
                # Pull out the q values and convert them
                u = []
                for fi in range(priors.n_filters):
                    q = [params[qX][0, fi, 0] for qX in priors.limb_dark_coeffs]
                    u.append(priors.ld_handler.convert_qtou(*q))
                return ln_likelihood + priors.ld_handler.ldtk_lnlike(
                    u, priors.limb_dark
                )
            else:
                return ln_likelihood

        #######################################################################
        #######################################################################

        # Now we can set up and run the sampler!
        sampler = NestedSampler(
            lnlike,
            prior_transform,
            n_dims,
            bound=bound,
            sample=sample,  # update_interval=float(n_dims),
            nlive=nlive,
            walks=walks,
            slices=slices,
        )
        """
        # Modification to include multiprocessing in single batches. 
        # If the n_procs is less than batches, then each batch gets additional 
        # processors to .
        if self.dynesty_procs>1:
            print(f"Running dynesty on {self.dynesty_procs} cores for this batch.")
            dynesty_pool = mp.Pool(self.dynesty_procs)
            sampler = NestedSampler(
                likelihood_calc.find_likelihood_parallel_processed,
                priors._convert_unit_cube,
                n_dims,
                bound=bound,
                sample=sample,  # update_interval=float(n_dims),
                nlive=nlive,
                walks=walks,
                slices=slices,
                pool=dynesty_pool,
                queue_size=self.dynesty_procs
            )
        
        else:
            sampler = NestedSampler(
                lnlike,
                prior_transform,
                n_dims,
                bound=bound,
                sample=sample,  # update_interval=float(n_dims),
                nlive=nlive,
                walks=walks,
                slices=slices,
            )"""

        try:
            sampler.run_nested(maxiter=maxiter, maxcall=maxcall, dlogz=dlogz)
        except BaseException as e:
            # Added for testing
            print(f"Exception ({type(e)}) encountered:")
            print(e)
            if type(e) == RuntimeError:
                print(
                    "If this is an error from sampling getting stuck, check your priors and the samples. This is often seen when the true value is at the edge of/outside your priors!"
                )

            print("Plotting samples at time of exception")
            # Plot the samples
            #results = sampler.results
            #results.best = results.samples[np.argmax(results.logl)]
            results = ResultsException(sampler)
            output_handler = OutputHandler(lightcurves, self._full_prior, self.host_r)
            output_handler._plot_samples(
                results, priors, "Exception_posteriors.png", plot_folder
            )
            raise
        #self.all_t0= likelihood_calc.all_t0
        # Pull out the results and calculate a few additional bits of info
        results = Results(sampler)
        """results = sampler.results

        # Normalise weights
        results.weights = get_normalised_weights(results)

        # Calculate covariance matrix and use to get uncertainties
        cov = get_covariance_matrix(results)
        diagonal = np.diag(cov)
        uncertainties = np.sqrt(diagonal)

        results.cov = cov
        results.uncertainties = uncertainties

        # Get the 16th and 84th percentiles to use as upper and lower errors
        # This is arguably better than using the covariances(???)
        median = np.median(results.samples, axis=0)
        per_16 = np.percentile(results.samples, 16, axis=0)
        per_84 = np.percentile(results.samples, 84, axis=0)

        results.median = median
        results.lower_err = abs(median - per_16)
        results.upper_err = abs(per_84 - median)
        results.per_16 = per_16
        results.per_84 = per_84

        # Save the best fit results for easy access
        results.best = results.samples[np.argmax(results.logl)]
        # results.best = median"""

        return results, n_dof

    def _run_full_retrieval(
        self,
        ld_fit_method,
        detrend,
        normalise,
        maxiter,
        maxcall,
        sample,
        nlive,
        dlogz,
        output_folder="./output_parameters",
        summary_file="summary_output.csv",
        full_output_file="full_output.csv",
        lightcurve_folder="./fitted_lightcurves",
        plot=True,
        plot_folder="./plots",
        marker_color="dimgrey",
        line_color="black",
        bound="multi",
        walks=100,
        slices=10,
    ):
        """
        Runs full retrieval with no folding/batching etc. Just a straight
        forward dynesty run.
        """
        priors, lightcurves = self._get_priors_and_curves(
            self.all_lightcurves,
            ld_fit_method,
            detrend=detrend,
            normalise=normalise,
            suppress_warnings=True,
        )

        # Set up output handler
        output_handler = OutputHandler(self.all_lightcurves, priors, self.host_r)

        # print(priors)
        results, ndof = self._run_dynesty(
            lightcurves,
            priors,
            maxiter,
            maxcall,
            sample,
            nlive,
            dlogz,
            bound,
            plot_folder,
            walks,
            slices,
        )
        return_results = deepcopy(results)
        # Print results to terminal
        try:
            print_results(results, priors, ndof)
        except Exception as e:
            print(e)

        # Save outputs parameters, plots, lightcurves!
        try:
            output_handler.save_results(
                [results],
                [priors],
                [lightcurves],
                output_folder,
                summary_file,
                full_output_file,
                plot_folder,
            )

        except Exception as e:
            print("Exception raised whilst saving results:")
            print(e)
            traceback.print_tb(e.__traceback__)
            print("Unable to save results")
        return return_results

    def _run_batched_retrieval(
        self,
        lightcurves,
        batches,
        ld_fit_method,
        detrend,
        normalise,
        maxiter,
        maxcall,
        sample,
        nlive,
        dlogz,
        full_return=False,
        folded=False,
        folded_P=None,
        folded_t0=None,
        output_folder="./output_parameters",
        summary_file="summary_output.csv",
        full_output_file="full_output.csv",
        lightcurve_folder="./fitted_lightcurves",
        plot=True,
        plot_folder="./plots",
        marker_color="dimgrey",
        line_color="black",
        bound="multi",
        filter_idx=None,
        walks=100,
        slices=10,
        n_procs=1,
    ):
        """
        Runs a retrieval using the given batches

        lightcurves : array_like, shape (n_telescopes, n_filters, n_epochs)
            The full lightcurves array to be retrieved.
        full_return : bool, optional
            If True will return all_results, all_priors, all_lightcurves.
            If False, will just return all_results. Default is False
        """
        # Get the full prior to make the output handler
        full_prior, full_lcs = self._get_priors_and_curves(
            lightcurves,
            ld_fit_method,
            None,
            detrend,
            normalise,
            folded,
            folded_P,
            folded_t0,
            suppress_warnings=True,
        )
        output_handler = OutputHandler(full_lcs, full_prior, self.host_r)

        n_batches = len(batches)

        mp_input = [
            [
                self,
                batch,
                bi,
                n_batches,
                lightcurves,
                output_handler,
                ld_fit_method,
                detrend,
                normalise,
                folded,
                folded_P,
                folded_t0,
                maxiter,
                maxcall,
                sample,
                nlive,
                dlogz,
                bound,
                plot_folder,
                walks,
                slices,
                output_folder,
                filter_idx,
                #summary_file,
                #full_output_file,
            ]
            for bi, batch in enumerate(batches)
        ]

        # with mp.Pool(processes=n_procs) as pool:
        # Cap n_procs so as not to exceed the number of cpus needed or the max number available, whichever is smaller
        # Modification to speed up processes when num-batches<n_procs.
        n_procs = np.amin([n_procs,mp.cpu_count()])
        # If the number of batches is smaller than the number of cores
        # to be used, it provides the additional cores to dynesty for 
        # multiprocessing
        if len(batches)<n_procs:
            self.dynesty_procs = n_procs//len(batches)
        else:
            self.dynesty_procs = 1
        n_procs = np.amin([n_procs,len(batches),mp.cpu_count()])
        if n_procs > 1:
            print('TransitFit is spawning n_procs = {:d} processes. Dynesty output might look scrambled!'.format(n_procs))

            pool = mp.Pool(n_procs,maxtasksperchild=1000)
            # Run the pool!
            # batch_run_results = [pool.map_async(_run_batch, i) for i in mp_input]
            batch_run_results = pool.map(_run_batch, mp_input)
            pool.close()
            pool.join()
            
        else: # For easy debugging, removing the case of multiprocessing when using only 1 CPU.
            batch_run_results=[_run_batch(mpinps) for mpinps in mp_input]

        all_results = np.array([r[0] for r in batch_run_results])
        all_priors = np.array([r[1] for r in batch_run_results])
        all_lightcurves = [r[2] for r in batch_run_results]

        # Make outputs etc
        if "filter" in os.path.basename(plot_folder):
            # plot_folder = os.path.dirname(plot_folder)
            plot_folder = os.path.join(
                os.path.dirname(plot_folder),
                "posteriors",
                os.path.basename(plot_folder),
            )
        else:
            plot_folder = os.path.join(plot_folder, "posteriors")

        output_handler.save_results(
            all_results,
            all_priors,
            all_lightcurves,
            output_folder,
            summary_file,
            full_output_file,
            plot_folder,
            folded,
        )

        if full_return:
            return all_results, all_priors, all_lightcurves

        return all_results

    def _run_folded_retrieval(
        self,
        ld_fit_method,
        detrend,
        normalise,
        maxiter,
        maxcall,
        sample,
        nlive,
        dlogz,
        output_folder="./output_parameters",
        summary_file="summary_output.csv",
        full_output_file="full_output.csv",
        lightcurve_folder="./fitted_lightcurves",
        plot=True,
        plot_folder="./plots",
        marker_color="dimgrey",
        line_color="black",
        max_parameters=25,
        overlap=2,
        bound="multi",
        walks=100,
        slices=10,
        n_procs=1,
    ):
        """
        For each filter, runs retrieval, then produces a phase-folded
        lightcurve. Then runs retrieval across wavelengths on the folded
        curves.
        """

        prefolded_batches = self._get_folding_batches(
            max_parameters, detrend, normalise, overlap
        )

        # Blank lists to fill with results etc
        results_list = []
        priors_list = []
        lightcurve_list = []

        print("Running pre-folding retrievals:")
        # Now run the batches for each filter
        for fi, filter_batches in enumerate(prefolded_batches):
            print("Filter {} of {}".format(fi + 1, self.n_filters))

            # Make a bunch of paths etc for saving the partial results
            filter_output_folder = os.path.join(
                output_folder, "filter_{}_parameters".format(fi)
            )
            filter_summary = "filter_{}_summary.csv".format(fi)
            filter_full_output = "filter_{}_full_output.csv".format(fi)
            filter_lightcurve_folder = os.path.join(
                lightcurve_folder, "filter_{}_curves".format(fi)
            )
            filter_plots_folder = os.path.join(
                plot_folder, "filter_{}_plots".format(fi)
            )

            results, priors, lightcurves = self._run_batched_retrieval(
                self.all_lightcurves,
                filter_batches,
                ld_fit_method,
                detrend,
                normalise,
                maxiter,
                maxcall,
                sample,
                nlive,
                dlogz,
                full_return=True,
                output_folder=filter_output_folder,
                summary_file=filter_summary,
                full_output_file=filter_full_output,
                lightcurve_folder=filter_lightcurve_folder,
                plot=plot,
                plot_folder=filter_plots_folder,
                marker_color=marker_color,
                line_color=line_color,
                bound=bound,
                filter_idx=fi,
                walks=walks,
                slices=slices,
                n_procs=n_procs,
            )

            results_list.append(results)
            priors_list.append(priors)
            lightcurve_list.append(lightcurves)

        # Now fold the lightcurves
        print("Folding lightcurves...")
        folded_curves, folded_P, folded_t0 = self._fold_lightcurves(
            results_list, priors_list, lightcurve_list
        )

        # Plot the folded lightcurves so we can check them
        for lci, lc in np.ndenumerate(folded_curves):
            quick_fname = "quick_plot-folded_curve_filter_{}.png".format(lci[1])
            quick_folder = os.path.join(plot_folder, "folded_curves/quicklook")
            quick_plot(lc, quick_fname, quick_folder, folded_t0, folded_P)

            print(f"Saving filter {lci[1]} quicklooks...")
            print(f"Phase plot saved to {os.path.join(quick_folder, quick_fname)}")

            # output the folded curves to file
            quicksave_path = os.path.join(
                lightcurve_folder,
                "quicklook",
                f"quick_plot-folded_curve_filter_{lci[1]}.csv",
            )
            lc.save(quicksave_path)
            print(f"Data file saved to {quicksave_path}")

        # Get the batches, and remember that now we are not detrending or
        # normalising since that was done in the first stage
        self.error_scaling = False
        folded_batches = self._get_non_folding_batches(
            folded_curves,
            max_parameters,
            ld_fit_method,
            detrend=False,
            normalise=False,
            overlap=overlap,
            error_scaling=False,
        )

        print("Running folded retrievals...")

        folded_results = self._run_batched_retrieval(
            folded_curves,
            folded_batches,
            ld_fit_method,
            False,
            False,
            maxiter,
            maxcall,
            sample,
            nlive,
            dlogz,
            False,
            True,
            folded_P,
            folded_t0,
            output_folder=output_folder,
            summary_file=summary_file,
            full_output_file=full_output_file,
            lightcurve_folder=lightcurve_folder,
            plot=plot,
            plot_folder=plot_folder,
            marker_color=marker_color,
            line_color=line_color,
            bound=bound,
            walks=walks,
            slices=slices,
            n_procs=n_procs,
        )

    def run_retrieval(
        self,
        ld_fit_method="independent",
        fitting_mode="auto",
        max_parameters=25,
        maxiter=None,
        maxcall=None,
        sample="auto",
        nlive=300,
        dlogz=None,
        plot=True,
        output_folder="./output_parameters",
        lightcurve_folder="./fitted_lightcurves",
        summary_file="summary_output.csv",
        full_output_file="full_output.csv",
        plot_folder="./plots",
        marker_color="dimgray",
        line_color="black",
        bound="multi",
        normalise=True,
        detrend=True,
        overlap=2,
        bin_data=True,
        cadence=2,
        binned_color="red",
        walks=100,
        slices=10,
        n_procs=1,
    ):
        """
        Runs dynesty on the data. Different modes exist and can be specified
        using the kwargs.

        Parameters
        ----------
        ld_fit_method : {`'coupled'`, `'single'`, `'independent'`, `'off'`}, optional
            Determines the mode of fitting of limb darkening parameters. The
            available modes are:
                - `'coupled'` : all limb darkening parameters are fitted
                  independently, but are coupled to a wavelength dependent
                  model based on the host parameters through `ldkt`
                - `'single'` : LD parameters are still tied to a model, but
                  only the first filter is actively fitted. The remaining
                  filters are estimated based off the ratios given by ldtk for
                  a host with the given parameters. This mode is useful for a
                  large number of filters, as `'coupled'` or `'independent'`
                  fitting will lead to much higher computation times.
                - `'independent'` : Each LD coefficient is fitted separately for
                  each filter, with no coupling to the ldtk models.
                - `'off'` : Will use the fixed value provided in the input file
            Default is `'independent'`
        fitting_mode : {'auto', 'all', 'folded', 'batched'}, optional
            Determines if the fitting algorithm is limited by max_parameters.
            If the number of parameters to be fitted exceeds max_parameters,
            then the retrieval will split into fitting each filter
            independently, phase-folding the detrended light curves to
            produce a single light curve for each filter and then fitting
            these phase-folded curves simultaneously. If fitting_mode is
            `'auto'`, then the mode used will be determined automatically.
            If fitting_mode is `'all'`, then all light curves will be
            attempted to be fitted simultaneously, regardless of the
            value of max_parameters. If fitting_mode
            is `'folded'`, then the folding approach will be used. Default is
            `'auto'`.
        max_parameters : int, optional
            The maximum number of parameters to use in a single retrieval.
            Default is 25.
        maxiter : int or `None`, optional
            The maximum number of iterations to run. If `None`, will
            continue until stopping criterion is reached. Default is `None`.
        maxcall : int or `None`, optional
            The maximum number of likelihood calls in retrieval. If None, will
            continue until stopping criterion is reached. Default is `None`.
        sample : str, optional
            Method used to sample uniformly within the likelihood constraint,
            conditioned on the provided bounds. Unique methods available are:
            uniform sampling within the bounds('unif'), random walks with fixed
            proposals ('rwalk'), random walks with variable (“staggering”)
            proposals ('rstagger'), multivariate slice sampling along preferred
            orientations ('slice'), “random” slice sampling along all
            orientations ('rslice'), “Hamiltonian” slices along random
            trajectories ('hslice'), and any callable function which follows
            the pattern of the sample methods defined in dynesty.sampling.
            'auto' selects the sampling method based on the dimensionality of
            the problem (from ndim). When ndim < 10, this defaults to 'unif'.
            When 10 <= ndim <= 20, this defaults to 'rwalk'. When ndim > 20,
            this defaults to 'hslice' if a gradient is provided and 'slice'
            otherwise. 'rstagger' and 'rslice' are provided as alternatives for
            'rwalk' and 'slice', respectively. Default is 'auto'.
        nlive : int, optional
            The number of live points to use in the nested sampling retrieval.
            Default is 300.
        dlogz : float, optional
            Retrieval iteration will stop when the estimated contribution of
            the remaining prior volume to the total evidence falls below this
            threshold. Explicitly, the stopping criterion is
            `ln(z + z_est) - ln(z) < dlogz`, where z is the current evidence
            from all saved samples and z_est is the estimated contribution from
            the remaining volume. The default is `1e-3 * (nlive - 1) + 0.01`.
        """
        
        # Auto mode detect
        if fitting_mode.lower() == "auto":
            # Generate the basic PriorInfo, needed to calculate how many
            # parameters would be fitted
            n_params_for_complete = self._calculate_n_params(
                self.all_lightcurves, None, ld_fit_method, normalise, detrend
            )

            if (
                n_params_for_complete > max_parameters
                and not self.n_total_lightcurves == 1
            ):
                # We have more than one lightcurve to fit - we can batch
                # If one filter has >= 3 epochs, we can do folded mode
                n_epochs_in_filter = np.sum(self.all_lightcurves != None, axis=(2, 0))

                if np.any(n_epochs_in_filter >= 3):
                    print("Auto mode detect has set 'folded' mode")
                    fitting_mode = "folded"
                else:
                    print("Auto mode detect has set 'batched' mode")
                    fitting_mode = "batched"
            else:
                print("Auto mode detect has set 'all' mode")
                fitting_mode = "all"

        if fitting_mode.lower() == "all":
            # In this mode, we are just running everything through a single
            # dynesty retrieval, not batching or anything.
            results = self._run_full_retrieval(
                ld_fit_method,
                detrend,
                normalise,
                maxiter,
                maxcall,
                sample,
                nlive,
                dlogz,
                output_folder,
                summary_file,
                full_output_file,
                lightcurve_folder,
                plot,
                plot_folder,
                marker_color,
                line_color,
                bound,
                walks,
                slices,
            )

        elif fitting_mode.lower() == "batched":
            # In this mode, we are generating batches which contain all
            # lightcurves for a filter. Batches may contain more than one
            # filter, but filters will not be split across multiple batches

            # Calculate the batches
            batches = self._get_non_folding_batches(
                self.all_lightcurves, max_parameters, detrend, normalise, overlap
            )

            results = self._run_batched_retrieval(
                self.all_lightcurves,
                batches,
                ld_fit_method,
                detrend,
                normalise,
                maxiter,
                maxcall,
                sample,
                nlive,
                dlogz,
                False,
                False,
                None,
                None,
                output_folder,
                summary_file,
                full_output_file,
                lightcurve_folder,
                plot,
                plot_folder,
                marker_color,
                line_color,
                bound,
                None,
                walks,
                slices,
                n_procs,
            )

        elif fitting_mode.lower() == "folded":
            # In this mode, we are running multi-epoch retrieval on each
            # filter separately, and then producing phase folded lightcurves
            # to run a multi-filter retrieval on.
            results = self._run_folded_retrieval(
                ld_fit_method,
                detrend,
                normalise,
                maxiter,
                maxcall,
                sample,
                nlive,
                dlogz,
                output_folder,
                summary_file,
                full_output_file,
                lightcurve_folder,
                plot,
                plot_folder,
                marker_color,
                line_color,
                max_parameters,
                overlap,
                bound,
                walks,
                slices,
                n_procs,
            )

        full_prior, _ = self._get_priors_and_curves(
            self.all_lightcurves,
            ld_fit_method,
            None,
            detrend,
            normalise,
            suppress_warnings=True,
        )

        output_handler = OutputHandler(self.all_lightcurves, full_prior, self.host_r)

        output_handler.save_complete_results(
            fitting_mode, full_prior, output_folder, summary_file
        )

        output_handler.save_final_light_curves(
            self.all_lightcurves, full_prior, lightcurve_folder
        )

        output_handler.plot_final_light_curves(
            self.all_lightcurves,
            full_prior,
            plot_folder,
            marker_color=marker_color,
            line_color=line_color,
            bin_data=bin_data,
            cadence=2,
            binned_color=binned_color,
            
        )

        if fitting_mode.lower() == "batched":
            print("'batched' mode was used to fit TransitFit lightcurves.")
            print("If this was the first run, it is suggested to rerun TransitFit for the model.")
            print("For wavelength independent parameters, use output from this run as priors.")

    ##########################################################
    #            PRIOR MANIPULATION                          #
    ##########################################################
    def _get_priors_and_curves(
        self,
        lightcurves,
        ld_fit_method,
        indices=None,
        detrend=True,
        normalise=True,
        folded=False,
        folded_P=None,
        folded_t0=None,
        suppress_warnings=False,
        #error_scaling=False,
    ):
        """
        Generates a prior info for a particular run:

        Parameters
        ----------
        lightcurves : array_like
            An array of the light curves which will be golbally considered for
            fitting.
        ld_fit_method : {'independent', 'single', 'coupled', 'off'}
            The mode to fit limb darkening coefficients with.
        indices : tuple or None
            If None, will fit all light curves. Otherwise, supply relevant
            indices of lightcurves to fit as a tuple:
            (telescope_indices, filter_indices, epoch_indices)
        detrend : bool, optional
            If True, will initialise detrending fitting. Default is True.
        normalise : bool, optional
            If True, will initialise normalisation fitting. Default is True.
        folded : bool, optional
            Set to True if using folded light curves (functionally only one
            epoch). Also turns off detrending and normalisation fitting.
            Default is False.
        folded_P : float, optional
            Required if folded is True. This is the period that the light
            curves are folded to
        folded_t0 : float, optional
            Required if folded is True. This is the t0 that the light curves
            are folded to. In the case where ttv mode is used (each epoch
            has been fitted to a different t0), then this should be the t0 of
            the first epoch (the one which everything else is folded back to)

        Returns
        -------
        priors : PriorInfo
            The fully initialised PriorInfo object
        lightcurves :
            The LightCurves in the correct format, with detrending and
            normalisation initialised

        """
        # Sort out indices and number of filters etc here.
        if indices is None:
            indices = tuple(np.array(list(np.ndindex(lightcurves.shape))).T)
        # indices = self._format_indices(indices)

        unique_indices = self._get_unique_indices(indices)

        # Indices of filters being used
        filter_indices = unique_indices[1]

        if folded:
            if filter_indices is None:
                raise ValueError("filter_indices must be provided for folded PriorInfo")

            lightcurve_subset = self._get_lightcurve_subset(lightcurves, indices)

            n_telescopes = 1
            n_filters = len(filter_indices)
            n_epochs = 1
            detrend = False
            normalise = False

        else:
            # get the lightcurve and detrending index array into the right shapes
            # and extract the relevant info.
            lightcurve_subset = self._get_lightcurve_subset(
                self.all_lightcurves, indices
            )
            detrending_indices = self._get_detrending_subset(indices)

            # Get unique indices and work out number of filters etc
            n_telescopes = len(unique_indices[0])
            n_filters = len(unique_indices[1])
            n_epochs = len(unique_indices[2])

        if self.host_r is not None:
            host_r = self.host_r[0]
        else:
            host_r = None

        # Set up the basic PriorInfo
        if type(self._prior_input) == str:
            # read in priors from a file
            priors = read_priors_file(
                self._prior_input,
                n_telescopes,
                n_filters,
                n_epochs,
                self.limb_darkening_model,
                filter_indices,
                folded,
                folded_P,
                folded_t0,
                host_r,
                self.allow_ttv,
                lightcurve_subset,
                suppress_warnings,
                self.error_scaling,
            )
        else:
            # Reading in from a list
            priors = parse_priors_list(
                self._prior_input,
                n_telescopes,
                n_filters,
                n_epochs,
                self.limb_darkening_model,
                filter_indices,
                folded,
                folded_P,
                folded_t0,
                host_r,
                self.allow_ttv,
                lightcurve_subset,
                suppress_warnings,
                self.error_scaling,
            )

        # Set up limb darkening
        if not ld_fit_method.lower() == "off":
            if ld_fit_method.lower() == "independent":
                priors.fit_limb_darkening(ld_fit_method)
            elif ld_fit_method.lower() in ["coupled", "single"]:
                if self._filter_input is None:
                    raise ValueError(
                        "filter_info must be provided for coupled and single ld_fit_methods"
                    )
                if self.host_T is None or self.host_z is None or self.host_logg is None:
                    raise ValueError(
                        "Filter info was provided but I am missing information on the host!"
                    )
                priors.fit_limb_darkening(
                    ld_fit_method,
                    self.host_T,
                    self.host_logg,
                    self.host_z,
                    self.filters[filter_indices],
                    self.n_ld_samples,
                    self.do_ld_mc,
                    self.ldtk_cache,
                    self.ldtk_uncertainty_multiplier,
                )
            else:
                raise ValueError(
                    "Unrecognised ld_fit_method '{}'".format(ld_fit_method)
                )

        if detrend:
            priors.fit_detrending(
                lightcurve_subset,
                self.detrending_info,
                detrending_indices,
                self.detrending_limits,
            )

        # Set up normalisation
        if normalise:
            priors.fit_normalisation(lightcurve_subset,self.normalise_limits)

        if self.error_scaling:
            priors.set_error_scaling(lightcurve_subset,self.scaling_limits, detrending_indices)

        return priors, lightcurve_subset

    ##########################################################
    #          LIGHTCURVE MANIPULATION                       #
    ##########################################################
    def _fold_lightcurves(self, results, priors, lightcurves):
        """
        Produces a set of folded lightcurves which can then be fitted across
        filters

        Parameters
        ----------
        results : array_like, shape (n_filters, )
            Each entry should be a list of results objects for each batch
        priors : array_like, shape (n_filters, )
            Each entry should be a list of PriorInfo objects for each batch
        lightcurves : array_like, shape (n_filters, )
            The lightcurves. Each entry should be a list of LightCurve arrays
            used for the batches within each filter

        Returns
        -------
        folded_lightcurves : np.array, shape (1, n_filters, 1)
            All the lightcurves, with each filter folded onto one epoch.
        folded_P : float
            The period that the lightcurves are folded with
        folded_t0 : float
            The t0 that all lightcurves are centred on.
        """
        # For each filter, go through the results and extract the best fit
        # values and uncertainties. For global values, we take the weighted
        # average and then use the detrending, normalisation, P and t0 values
        # to produce a single LightCurve comprised by folding all the
        # detrended and normalised curves within a filter
        n_filters = self.n_filters
        allow_ttv = self.allow_ttv

        ###############################################################
        ###                  GET P AND t0 VALUES                    ###
        ###############################################################
        result_handler = OutputHandler(
            self.all_lightcurves, self._full_prior, self.host_r
        )

        results_dicts = []
        for fi in range(n_filters):
            n_batches = len(results[fi])
            for bi in range(n_batches):
                result_dict = result_handler.get_results_dict(
                    results[fi][bi], priors[fi][bi], lightcurves[fi][bi]
                )
                results_dicts.append(result_dict)

        combined_results = result_handler.combine_results_dicts(results_dicts)

        # Get best P value
        if self.allow_ttv:
            # P is fixed
            best_P, best_P_err = combined_results["P"][None, None, None][0, 0], 0
        else:
            best_P, best_P_err = weighted_avg_and_std(
                combined_results["P"][None, None, None][:, 0],
                combined_results["P"][None, None, None][:, -1],
                single_val=True,
            )

        # Get best t0 values, allowing for ttv mode
        best_t0, best_t0_err = np.full(self.n_epochs, None), np.full(
            self.n_epochs, None
        )

        for ei in range(self.n_epochs):
            best_t0[ei], best_t0_err[ei] = weighted_avg_and_std(
                combined_results["t0"][None, None, ei][:, 0],
                combined_results["t0"][None, None, ei][:, -1],
                single_val=True,
            )

        print("Folding light curves with these parameters:")
        print("P = {} ± {}".format(round(best_P, 8), round(best_P_err, 8)))
        if self.allow_ttv:
            for i, t0i in enumerate(best_t0):
                print(
                    "t0 = {} ± {} (epoch {})".format(
                        round(t0i, 6), round(best_t0_err[i], 6), i
                    )
                )
        else:
            print("t0 = {} ± {}".format(round(best_t0[0], 6), round(best_t0_err[0], 6)))

        ###############################################################
        ###            NORMALISATION/DETRENDING/FOLDING             ###
        ###############################################################
        # Updated version for v2.3
        final_batched_lightcurves = [[] for i in range(self.n_filters)]

        # loop over each lightcurve, using the combined results dict
        for li, lc in np.ndenumerate(self.all_lightcurves):
            if lc is not None:
                if self._full_prior.detrend:
                    # Get the detrending coeffs
                    method_idx = lc.detrending_method_idx
                    # Get the detrending coeffs names for this lc
                    detrending_coeffs = self._full_prior.detrending_coeffs[method_idx]

                    # Now get the weighted best values for them!
                    best_d = []
                    for d in detrending_coeffs:
                        best_d.append(
                            weighted_avg_and_std(
                                combined_results[d][li][:, 0],
                                combined_results[d][li][:, -1],
                                single_val=True,
                            )[0]
                        )
                    #best_d_new=best_d.append(best_t0[li[2]])
                    #best_d_new=[*best_d,*best_t0[li[2]]]

                    #print(best_d,'A',best_d_new)
                else:
                    best_d = None
                    #best_d_new=[best_d,*best_t0[li[2]]]
                    #print(best_d,'B',best_d_new)
                
                try:    
                    best_d_new=[*best_d,best_t0[li[2]],best_P]
                except TypeError:
                    best_d_new=[best_d,best_t0[li[2]],best_P]

                if self._full_prior.normalise:
                    norm = weighted_avg_and_std(
                        combined_results["norm"][li][:, 0],
                        combined_results["norm"][li][:, -1],
                        single_val=True,
                    )[0]
                else:
                    norm = 1
                
                if self._full_prior.error_scaling:
                    escale = weighted_avg_and_std(
                        combined_results["escale"][li][:, 0],
                        combined_results["escale"][li][:, -1],
                        single_val=True,
                    )[0]
                else:
                    escale = None 

                # Make the detrended light curve and fold to the final t0
                #detrended_curve = lc.create_detrended_LightCurve(best_d, norm)
                detrended_curve = lc.create_detrended_LightCurve(best_d_new, norm, escale)

                # Fold the curve using the best t0 for the epoch and P
                # We are folding each curve to be centred on best_t0[0]
                folded_curve = detrended_curve.fold(best_t0[li[2]], best_P, best_t0[0])
                final_batched_lightcurves[li[1]].append(folded_curve)

        # Now we go through detrended and folded lightcurve, and combine them
        # into one lightcurve per filter
        final_lightcurves = []
        for fi, filter_curves in enumerate(final_batched_lightcurves):
            # Go through each filter and combine the curves!
            if len(filter_curves) == 1:
                # No need to combine
                final_lightcurves.append(filter_curves[0])
            else:
                combined_curve = filter_curves[0].combine(
                    filter_curves[1:], filter_idx=fi
                )
                # Need to loop and combine
                final_lightcurves.append(combined_curve)

        return (
            np.array(final_lightcurves).reshape(1, self.n_filters, 1),
            best_P,
            best_t0[0],
        )

        """
        # Now we do the detrending stuff and fold the lightcurves
        # Remember that each batch will only contain one filter

        final_batched_lightcurves = [[] for i in range(self.n_filters)]
        #final_batched_lightcurves = []
        for fi, filter_results in enumerate(results):
            # Loop through each filter
            for ri, result in enumerate(filter_results):
                # Loop through each batch within a filter

                # Get the batch-relevant prior, lightcurves and results dicts
                prior = priors[fi][ri]
                lcs = lightcurves[fi][ri]

                results_dict, errors_dict = prior._interpret_final_results(result)

                # Loop through lightcurves and detrend/normalise
                for i, lc in np.ndenumerate(lcs):
                    if lc is not None:
                        # Get some sub indices.
                        tidx, fidx, eidx = i
                        if prior.detrend:
                            method_idx = lc.detrending_method_idx
                            # Get the detrending coeffs for this lc
                            detrending_coeffs = prior.detrending_coeffs[method_idx]

                            best_d = [results_dict[d][i] for d in detrending_coeffs]
                        else:
                            best_d = None

                        norm = results_dict['norm'][i]

                        # Make the detrended light curve and fold to the final t0
                        detrended_curve = lc.create_detrended_LightCurve(best_d, norm)

                        # Fold the curve using the best t0 for the epoch and P
                        # We are folding each curve to be centred on best_t0[0]
                        folded_curve = detrended_curve.fold(best_t0[eidx], best_P, best_t0[0])
                        final_batched_lightcurves[fi].append(folded_curve)

        # Now we go through detrended and folded lightcurve, and combine them
        # into one lightcurve per filter
        final_lightcurves = []
        for fi, filter_curves in enumerate(final_batched_lightcurves):
            # Go through each filter and combine the curves!
            if len(filter_curves) == 1:
                # No need to combine
                final_lightcurves.append(filter_curves[0])
            else:
                combined_curve = filter_curves[0].combine(filter_curves[1:], filter_idx=fi)
                # Need to loop and combine
                final_lightcurves.append(combined_curve)

        return np.array(final_lightcurves).reshape(1, self.n_filters, 1), best_P, best_t0[0]
        """

    def _get_lightcurve_subset(self, lightcurves, indices):
        """
        Pulls out the subset of lightcurves given by the indices.
        """
        # Get the unique indices
        unique_indices = self._get_unique_indices(indices)
        # Make some empty arrays for us to populate
        lightcurves_subset = np.full(tuple(len(idx) for idx in unique_indices), None)

        # Go through each index being used and put in the relevant info
        # deepcopy is used here to ensure we don't end up with clashing
        # attributes (e.g. from detrending twice)
        for index in np.array(indices).T:
            subset_index = self._full_to_subset_index(indices, index)
            lightcurves_subset[subset_index] = deepcopy(lightcurves[tuple(index)])

        return lightcurves_subset

    def _get_detrending_subset(self, indices):
        """
        Pulls out the detrending indices for the given
        telescope, filter and epoch indices (given as tuple), assuming that we
        are working on all input lightcurves, not a folded version (though that
        shouldn't be using this function)
        """
        # Get the unique indices
        unique_indices = self._get_unique_indices(indices)

        detrending_indices = np.full(tuple(len(idx) for idx in unique_indices), None)

        # Go through each index being used and put in the relevant info
        # deepcopy is used here to ensure we don't end up with clashing
        # attributes (e.g. from detrending twice)
        for index in np.array(indices).T:
            subset_index = self._full_to_subset_index(indices, index)
            detrending_indices[subset_index] = deepcopy(
                self.detrending_index_array[tuple(index)]
            )

        return detrending_indices

    ##########################################################
    #        FUNCTIONS FOR BATCHING ETC                      #
    ##########################################################
    def _get_folding_batches(
        self,
        max_parameters,
        ld_fit_method,
        detrend,
        normalise,
        overlap=2,
        random_order=True,
    ):
        """
        Splits all_lightcurves into single-filter, multi-epoch batches
        to be fitted, which will allow us to produce folded lightcurves. This
        includes the option to have batches overlapping so that they share some
        info.

        Parameters
        ----------
        max_parameters : int
            The maximum number of parameters to have in a single batch
        ld_fit_method : str
            The limb darkening fit method
        detrend : bool
            If True, detrending will be used
        normalise : bool
            If true, normalisation will be used
        overlap : int, optional
            The number of epochs to overlap in each batch. This will be adhered
            to where possible. Default is 2.
        random_order : bool, optional
            If True, will shuffle the epochs before batching to reduce
            correlations from grouping them. Default is True

        Returns
        -------
        batches : array_like, shape (n_filters,)
            Each entry in the array is a list of batches for the given filter.
        """
        # All of the batches. Each entry will be the batches for a filter
        all_batches = []

        # Loop through each filter
        for fi in range(self.n_filters):
            # All the batches for this filter
            filter_batches = []

            # Find the indices of the lightcurves within the filter
            indices = np.where(self.all_lightcurves[:, fi, :] != None)
            n_curves = len(indices[0])
            # Randomise the order to minimise correlations
            if random_order:
                random_indices = np.arange(n_curves)
                np.random.shuffle(random_indices)
                indices = (indices[0][random_indices], indices[1][random_indices])

            # Put the fi index back into indices tuple
            indices = (indices[0], np.full(n_curves, fi), indices[1])

            # Loop through each lightcurve to get the batches!
            # What index are we starting the loop from? This is used to get the
            # overlapping batches
            start_idx = 0

            # Flag to check if we have got batches containing the full range
            # of epoch/filter
            done = n_curves == 0

            while not done:
                # A single batch - telescope, filter, epoch indices
                single_batch = ([], [], [])
                for i in range(start_idx, n_curves):
                    idx = (indices[0][i], indices[1][i], indices[2][i])
                    # Make the test batch by appending this lightcurve to the
                    # current single_batch
                    test_batch = tuple((single_batch[j] + [idx[j]] for j in range(3)))
                    # Get the number of parameters for the test_batch
                    n_params = self._calculate_n_params(
                        self.all_lightcurves,
                        test_batch,
                        ld_fit_method,
                        normalise,
                        detrend,
                    )

                    if n_params <= max_parameters:
                        # Add the lightcurve to the single batch and carry on
                        single_batch = deepcopy(test_batch)

                        if i == (n_curves - 1):
                            # We have completed the loop - we have all the
                            # lightcurves in at least one batch
                            done = True

                            # Save this final batch
                            filter_batches.append(single_batch)

                    else:
                        # Is this the only lightcurve in the batch?
                        # If so, we just want to add this curve and move on
                        if len(test_batch[0]) == 1:
                            single_batch = deepcopy(test_batch)
                            start_idx = i + 1

                        else:
                            # We will exlude the last added lightcurve and work
                            # only with the current confirmed batch

                            # Do we have enough to ensure an overlap?
                            if len(single_batch[0]) > overlap:
                                start_idx = i - overlap

                            else:
                                print(
                                    "Unable to ensure overlap of {} between batch {} and {} for filter {}".format(
                                        overlap,
                                        len(filter_batches),
                                        len(filter_batches) + 1,
                                        fi,
                                    )
                                )

                                if len(single_batch[0]) > 1:
                                    # We can at least try for an overlap of 1
                                    start_idx = i - 1

                                else:
                                    # We just can't overlap
                                    start_idx = i

                        # Save the batch to start a new one
                        filter_batches.append(single_batch)

                        # Now check to see if we have done all the lightcurves
                        # for the filter
                        done = start_idx == n_curves
                        break

            all_batches.append(filter_batches)
        return all_batches

    def _get_non_folding_batches(
        self, lightcurves, max_parameters, ld_fit_method, detrend, normalise, overlap=2, error_scaling=False,
    ):
        """
        Splits lightcurves into batches by filter, attempting to ensure that
        batches do not require more than max_parameters to fit. Each batch will
        contain every light curve in the filter, which trumps max_parameters.
        Where possible, the filter batches will overlap, allowing each batch
        to share information on some level.

        Parameters
        ----------
        lightcurves : array_like, shape (n_telescopes, n_filters, n_epochs)
            The LightCurves
        max_parameters : int
            The maximum number of parameters to have in a single batch
        ld_fit_method : str
            The limb darkening fit method
        detrend : bool
            If True, detrending will be used
        normalise : bool
            If true, normalisation will be used
        overlap : int, optional
            The number of epochs to overlap in each batch. This will be adhered
            to where possible. Default is 2.
        lightcurves : array_like, shape (n_telescopes, n_filters, n_epochs)
            The lightcurves to be batched.

        Returns
        -------
        batches : array_like, shape (n_batches)
            The final batches. Each entry is a tuple of
            (telescope indices, filter indices, epoch indices) using the
            indices of lightcurves. Each of these batches can then be passed to
            _get_priors_and_curves.
        """
        # Get some info out of the lightcurves
        n_telescopes, n_filters, n_epochs = lightcurves.shape

        # The final batches
        all_batches = []

        # What filter are we starting the loop from? This is used to get the
        # overlapping batches
        start_idx = 0

        # Flag to check if each filter is in at least one batch
        done = False

        while not done:
            # A single batch - telescope, filter, epoch indices
            single_batch = ([], [], [])

            filters_in_batch = 0
            # loop through each filter
            for fi in range(start_idx, n_filters):

                # For each filter, we need to pull out all the light curves
                # and calculate how many parameters we would be fitting if
                # this filter were added to the batch.

                # First, get the indices of lightcurves in the filter:
                indices = np.where(lightcurves[:, fi, :] != None)
                n_curves = len(indices[0])

                # Put the fi index back into indices tuple, converting to list
                indices = (
                    list(indices[0]),
                    list(np.full(n_curves, fi)),
                    list(indices[1]),
                )

                # Make the test batch
                test_batch = tuple((single_batch[j] + indices[j] for j in range(3)))

                # Get the number of params for the test batch
                n_params = self._calculate_n_params(
                    lightcurves, test_batch, ld_fit_method, normalise, detrend
                )

                if n_params <= max_parameters:
                    # Add the filter to the batch
                    single_batch = deepcopy(test_batch)
                    filters_in_batch += 1
                    # Check to see if we have completed the loop and have all
                    # the filters in at least one batch
                    if fi == (self.n_filters - 1):
                        done = True
                        # save the final batch
                        all_batches.append(single_batch)

                else:
                    # Too many params - what to do?
                    # Is this the only filter in the batch? If so, add it and
                    # move on as there's too many parameters to put this
                    # filter in with anything else
                    if filters_in_batch == 0:
                        filters_in_batch += 1
                        single_batch = deepcopy(test_batch)

                        start_idx = fi + 1

                    else:
                        # Exclude the last added filter - don't update
                        # single_batch

                        # Do we have enough to ensure an overlap?
                        if filters_in_batch > overlap:
                            start_idx = fi - overlap

                        else:
                            print(
                                "Unable to ensure overlap of {} between batch {} and batch {}".format(
                                    overlap, len(all_batches), len(all_batches) + 1
                                )
                            )

                            # Can we at least try to get an overlap of 1?
                            if filters_in_batch > 1:
                                print("Attempting overlap of 1")
                                start_idx = fi - 1
                            else:
                                print("Overlap cannot be made")
                                # We just can't overlap. Ah well
                                start_idx = fi

                    # save the batch and start a new one
                    all_batches.append(single_batch)

                    # Now check to see if all the filters are included in
                    # at least one batch
                    done = start_idx == n_filters
                    break

        return all_batches

    ##########################################################
    #             UTILITY FUNCTIONS                          #
    ##########################################################
    def _calculate_n_params(
        self, lightcurves, indices, ld_fit_method, normalise, detrend
    ):
        """
        Calculates the number of parameters which would be fitted for a
        given set of filter and epoch indices

        This function exists because it's much faster than repeatedly making
        PriorInfos for different combos.

        Parameters
        ----------
        indices : tuple
            The tuple of indices to consider. Must be given as
            (telescope_indices, filter_indices, epoch_indices)

        """
        indices = self._format_indices(indices)

        lightcurve_subset = self._get_lightcurve_subset(lightcurves, indices)
        detrending_indices = self._get_detrending_subset(indices)

        unique_indices = self._get_unique_indices(indices)

        n_filters, n_epochs = (len(unique_indices[1]), len(unique_indices[2]))

        n_lightcurves = (lightcurve_subset != None).sum()

        # Account for global parameters
        n_params = self.n_global_params

        # Account for ttv
        if self.allow_ttv:
            # t0 for each epoch
            n_params += n_epochs

        # Account for filter-specific parameters - rp and LD coeffs
        if ld_fit_method in ["independent", "coupled"]:
            n_params += n_filters * (1 + self.n_ld_params)
        else:  # single fitting mode being used
            n_params += n_filters + self.n_ld_params

        # Account for normalisation
        if normalise:
            n_params += n_lightcurves
        
        if self.error_scaling:
            n_params += n_lightcurves
        # Account for detrending

        if detrend:
            for i in np.array(indices).T:
                subset_i = self._full_to_subset_index(indices, i)
                if lightcurve_subset[subset_i][0] is not None:
                    detrending_index = detrending_indices[subset_i][0]
                    detrending_info = self.detrending_info[detrending_index]
                    detrending_model = self.detrending_models[detrending_index]
                    if detrending_info[0] in ["nth order", "custom"]:
                        n_params += detrending_model.n_params
                    elif detrending_info[0] == "off":
                        pass
                    else:
                        raise ValueError(
                            "Unrecognised detrending model {}".format(
                                detrending_info[0]
                            )
                        )

        return n_params

    def _format_indices(self, indices):
        """
        If passed a set of indices, checks they are usable.
        If indices is None, sets them to cover all of the possible values
        """
        if indices is None:
            return tuple(np.array(list(np.ndindex(self.all_lightcurves.shape))).T)

        return indices

    def _get_unique_indices(self, indices):
        """
        When given a tuple of indices of all light curves to consider,
        gets all the unique values
        """
        return [np.unique(i) for i in indices]

    def _full_to_subset_index(self, subset_indices, full_index):
        """
        Converts an index which uses the notation of full parameter space to
        a subset. USeful for converting between overall indexing and indexing
        within a batch

        Parameters
        ----------
        subset_indices : tuple
            The indices which define the full subset of light curves
        full_index : array_like, shape (3, )
            The full-notation index to be converted
        """
        unique_indices = self._get_unique_indices(subset_indices)

        return tuple(
            (np.where(unique_indices[i] == full_index[i])[0])
            for i in range(len(full_index))
        )

    def _subset_to_full_index(self, subset_indices, subset_index):
        """
        Converts an index from notation within a batch to the full indexing.
        Inverse of _full_to_subset_index

        Parameters
        ----------
        subset_indices : tuple
            The indices which define the full subset of light curves
        subset_index : array_like, shape (3, )
            The subset-notation index to be converted
        """
        unique_indices = self._get_unique_indices(subset_indices)

        return unique_indices[subset_index]


# External definition of running retrieval on a batch. Exists for
# parallelisation purposes.
def _run_batch(x):
    """
    Subprocess to run a batch
    """
    (
        retriever,
        batch,
        bi,
        n_batches,
        lightcurves,
        output_handler,
        ld_fit_method,
        detrend,
        normalise,
        folded,
        folded_P,
        folded_t0,
        maxiter,
        maxcall,
        sample,
        nlive,
        dlogz,
        bound,
        plot_folder,
        walks,
        slices,
        output_folder,
        filter_idx,
    ) = x

    print("Running batch {} of {}".format(bi + 1, n_batches))
    # Now we want to get the lightcurves and priors for each batch
    batch_prior, batch_lightcurves = retriever._get_priors_and_curves(
        lightcurves,
        ld_fit_method,
        batch,
        detrend,
        normalise,
        folded,
        folded_P,
        folded_t0,
        suppress_warnings=True,
    )

    # Run the retrieval!
    results, ndof = retriever._run_dynesty(
        batch_lightcurves,
        batch_prior,
        maxiter,
        maxcall,
        sample,
        nlive,
        dlogz,
        bound,
        plot_folder,
        walks,
        slices,
    )

    output_handler._quicksave_result(
        results, batch_prior, batch_lightcurves, output_folder, filter_idx, bi
    )

    #print('The shapes are:',np.shape(results),np.shape(batch_prior),np.shape(batch_lightcurves))
    return results, batch_prior, batch_lightcurves



'''
Class to handle limb darkening

'''
import numpy as np
from ._ldtk_handler import LDTKHandler

class LimbDarkeningHandler:
    '''
    The LimbDarkeningHandler is designed to convert fitting parameters in
    the range [0,1] and convert them into physically allowed values for
    limb darkening coefficients for different models. This conversion is
    based on Kipping 2013 https://arxiv.org/abs/1308.0009.

    Parameters
    ----------
    default_model : str
        The default limb darkening model to use. Accepted values are
            - 'linear'
            - 'quadratic'
            - 'squareroot'
            - 'power2'
            - 'nonlinear'
        This model will be the default for conversions unless otheriwse
        specified.
    low_lim : float, optional
        The lower limit to use in conversion in the case where there are
        open bounds on a coefficient (power2 and nonlinear models). Note
        that in order to conserve sampling density in all regions for the
        power2 model, you should set lower_lim=-high_lim. Default is -5
    high_lim : float, optional
        The upper limit to use in conversion in the case where there are
        open bounds on a coefficient (power2 and nonlinear models). Note
        that in order to conserve sampling density in all regions for the
        power2 model, you should set lower_lim=-high_lim. Default is 5
    '''

    def __init__(self, default_model, low_lim=-5, high_lim=5):
        if default_model not in ['linear','quadratic','squareroot','power2','nonlinear']:
            raise ValueError('Unrecognised limb darkening model {}'.format(default_model))

        self.default_model = default_model

        # the default high and low bounds to use for unconstrained values
        self.high = high_lim
        self.low = low_lim

        self.ldtk_handler = None

        self.host_T = None
        self.host_logg = None
        self.host_z = None

    def get_required_coefficients(self, model=None):
        '''
        Finds the number of coefficients required for a given model and returns
        a string name for each of the form 'uX' where X is a number.

        Parameters
        ----------
        model : str or None, optional
            Model to get the coefficients for. Default is self.default_model

        Returns
        -------
        coeficient names : list of str
            The coefficient names
        '''
        if model is None:
            model = self.default_model

        if model == 'linear':
            return ['q0']
        if model in ['quadratic','squareroot', 'power2']:
            return ['q0', 'q1']
        if model == 'nonlinear':
            return ['q0', 'q1', 'q2', 'q3']

        raise ValueError('Unrecognised model {}'.format(model))

    def convert_qtou(self, *q, model=None):
        '''
        Takes parameters q distributed between [0,1] and converts them into
        physical values for limb darkening coefficients.

        Conversions for quadratic, square root, and logarithmic are from
        Kipping 2013 https://arxiv.org/abs/1308.0009 for two-parameter limb
        darkening methods

        Notes
        -----
        This is the inverse of convert_utoq
        '''
        if model is None:
            model = self.default_model

        if model == 'linear':
            # coefficient is limited to 0<A<1 by Kipping criteria
            return [q[0]]

        if model == 'quadratic':
            return 2 * np.sqrt(q[0]) * q[1], np.sqrt(q[0]) * (1 - 2 * q[1])

        if model == 'squareroot':
            return np.sqrt(q[0]) * (1 - 2 * q[1]), 2 * np.sqrt(q[0]) * q[1]

        if model == 'power2':
            # This parameterisation has been derived for TransitFit and is
            # explained in the paper
            u1 = q[0] * (1 - self.low) + self.low

            if u1 < 0:
                # negative quadrant
                return u1, self.low * (1 - q[1])

            else:
                # positive quadrant
                return u1, self.high * q[1]

        if model == 'nonlinear':
            # This is an 'outstanding and formidable problem' to apply the
            # Kipping (2013) method to (see Kipping 2016) and as such, we can
            # only fit for free parameters, without reparameterising to allow
            # only physically valid combinations.
            u1 = q[0] * (self.high - self.low) + self.low
            u2 = q[1] * (self.high - self.low) + self.low
            u3 = q[2] * (self.high - self.low) + self.low
            u4 = q[3] * (self.high - self.low) + self.low

            return u1, u2, u3, u4

        raise ValueError('Unrecognised model {}'.format(model))

    def convert_utoq(self, *u, model=None):
        '''
        Takes actual values of the LD coefficients and converts them to a
        value q between [0,1].

        Conversions for quadratic, square root, and logarithmic are from
        Kipping 2013 https://arxiv.org/abs/1308.0009 for two-parameter limb
        darkening methods

        Notes
        -----
        This is the inverse of convert_qtou
        '''
        if model is None:
            model = self.default_model

        if model == 'linear':
            # coefficient is limited to 0<A<1 by Kipping criteria
            return [q[0]]

        if model == 'quadratic':
            return (u[0] + u[1]) ** 2, u[0]/(2 * (u[0] + u[1]))

        if model == 'squareroot':
            return (u[0] + u[1]) ** 2, u[1]/(2 * (u[0] + u[1]))

        if model == 'power2':
            # This parameterisation has been derived for TransitFit and is
            # explained in the paper
            q1 = (u[0] - self.low)/(1 - self.low)

            if u[0] < 0:
                # negative quadrant
                return u1, 1 - (u[1]/self.low)

            else:
                # positive quadrant
                return u1, u[1]/self.high

        if model == 'nonlinear':
            # This is an 'outstanding and formidable problem' to apply the
            # Kipping (2013) method to (see Kipping 2016) and as such, we can
            # only fit for free parameters, without reparameterising to allow
            # only physically valid combinations.
            q0 = (u[0] - self.low) /  (self.high - self.low)
            q1 = (u[1] - self.low) /  (self.high - self.low)
            q2 = (u[2] - self.low) /  (self.high - self.low)
            q3 = (u[3] - self.low) /  (self.high - self.low)

            return q0, q1, q2, q3

        raise ValueError('Unrecognised model {}'.format(model))

    def convert_qtou_with_errors(self, q, q_err, model=None):
        '''
        Converts kipping parameters with errors into physical LDCs with errors
        '''
        if model is None:
            model = self.default_model

        u = list(self.convert_qtou(*q, ))

        if model == 'linear':
            u_err = q_err

            return [u], [u_err]

        if model  == 'quadratic':
            u_err = []
            u_err.append(np.sqrt(((q[1] * q_err[0])**2)/q[0] + 4 * q[0] * q_err[1]**2))
            u_err.append(np.sqrt((((1- 2 * q[1]) * q_err[0])**2)/(0.25 * q[0]) + 4 * q[0] * q_err[1] ** 2))

            return u, u_err

        if model == 'squareroot':
            u_err = []
            u_err.append(np.sqrt((((1- 2 * q[1]) * q_err[0])**2)/(0.25 * q[0]) + 4 * q[0] * q_err[1] ** 2))
            u_err.append(np.sqrt(((q[1] * q_err[0])**2)/q[0] + 4 * q[0] * q_err[1]**2))

            return u, u_err

        if model == 'power2':
            u_err = []
            u_err.append(abs(1-self.low) * q_err[0])
            if u[1] < 0:
                u_err.append(abs(self.low) * q_err[1])
            else:
                u_err.append(abs(self.high) * q_err[1])

            return u, u_err

        if model == 'nonlinear':
            u_err = [abs(self.high - self.low) * err for err in q_err]

            return u, u_err


    def initialise_ldtk(self, host_T, host_logg, host_z, filters,
                        model=None, n_samples=20000, do_mc=False,
                        cache_path=None, ldtk_uncertainty_multiplier=1.):
        '''
        Sets up an LDTKHandler to deal with interfacing between ldtk and
        TransitFit

        Parameters
        ----------
        host_T : tuple
            The effective temperature of the host star, in Kelvin, given as a
            (value, uncertainty) pair.
        host_logg : tuple
            The log_10 of the surface gravity of the host star, with gravity
            measured in cm/s2. Should be given as a (value, uncertainty) pair.
        host_z : tuple
            The metalicity of the host, given as a (value, uncertainty) pair.
        filters : array_like
            The set of filters, given in [low, high] limits for the wavelengths
            with the wavelengths given in nanometers. The ordering of the
            filters should correspond to the filter_idx parameter used
            elsewhere.
        ld_method : str or None, optional
            The model of limb darkening to use. Allowed values are 'linear',
            'quadratic', 'squareroot', 'power2', and 'nonlinear'. If None, will
            default to self.default_model. Default is None.
        n_samples : int, optional
            The number of limb darkening profiles to create. Passed to
            ldtk.LDPSetCreator.create_profiles(). Default is 20000.
        do_mc : bool, optional
            If True, will use MCMC to estimate coefficient uncertainties more
            accurately. Default is False.
        cache_path : str, optional
            This is the path to cache LDTK files to. If not specified, will
            default to the LDTK default
        ldtk_uncertainty_multiplier: float, optional
            (From LDTK:) The uncertainty multiplier ϵ is a subjective factor
            that defines how strongly the LD profile (or the prior created
            from it) constrains the final analysis (that is, how much we
            trust the stellar atmosphere models used to create the profiles.)
        '''
        if model is None:
            model = self.default_model

        self.host_T = host_T
        self.host_logg = host_logg
        self.host_z = host_z

        self.ldtk_handler = LDTKHandler(host_T, host_logg, host_z, filters,
                                        model, n_samples, do_mc, cache_path,
                                        ldtk_uncertainty_multiplier)


    def ldtk_lnlike(self, coeffs, model=None):
        '''
        Uses LDTK to evaluate the log likelihood for a set of coefficients.
        Note that coeffs should be the values returned by convert_qtoA
        and NOT the unit values.

        Parameters
        ----------
        coeffs : array_like, shape (n_filters, n_coeffs)
            The coefficients to evaluate the log likelihood for.
        ld_model : str, optional
            The model to use. Defaults to self.default_model

        Returns
        -------
        lnlike : float
            The log likelihood of the coefficients
        '''

        if self.ldtk_handler is None:
            raise Exception('ldtk_handler is not initialised. Please initialise using initialise_ldtk()')

        if model is None:
            model = self.default_model

        return self.ldtk_handler.lnlike(coeffs, model)

    def ldtk_estimate(self, ld0_values, model=None):
        '''
        Uses LDTK to estimate the LDCs for all filters when given a set of LDCs
        for filter 0, based on the ratios between the best values
        found in initialisation.

        Parameters
        ----------
        ld1_values : float or array_like
            The LD parameters for filter 0
        ld_model : str, optional
            The limb darkening model to use. Defaults to self.default_model

        Returns
        -------
        all_ld_values : array_like, shape (n_filters, n_coeffs)
            The estimated limb darkening parameters
        '''
        if self.ldtk_handler is None:
            raise Exception('ldtk_handler is not initialised. Please initialise using initialise_ldtk()')

        if model is None:
            model = self.default_model

        return self.ldtk_handler.estimate_values(ld0_values, model)
